from __future__ import division

import copy
import gzip
import logging
import multiprocessing as mp
import os
import threading
import uuid
from abc import abstractmethod, ABCMeta
from bisect import bisect_right
from builtins import object
from collections import defaultdict
from queue import Empty
from timeit import default_timer as timer
import tempfile
import shutil
from datetime import datetime

import msgpack
import numpy as np
import pysam
import tables as t
from future.utils import with_metaclass, viewitems, string_types

from genomic_regions import GenomicRegion, RegionBased
from .config import config
from .general import MaskFilter, Mask
from .hic import Hic
from .matrix import Edge, RegionPairsTable
from .regions import genome_regions
from .tools.general import RareUpdateProgressBar, add_dict, find_alignment_match_positions, WorkerMonitor
from .tools.sambam import natural_cmp
from .tools.files import split_sam_pairs


logger = logging.getLogger(__name__)


def generate_pairs(sam1_file, sam2_file, regions,
                   restriction_enzyme=None, output_file=None,
                   read_filters=(), check_sorted=True,
                   threads=1, batch_size=10000000):
    """
    Generate Pairs object from SAM/BAM files.

    This is a convenience function that let's you create a
    Pairs object from SAM/BAM data in a single step.

    :param sam1_file: Path to a sorted SAM/BAM file (1st mate)
    :param sam2_file: Path to a sorted SAM/BAM file (2nd mate)
    :param regions: Path to file with restriction fragments (BED, GFF)
                    or FASTA with genomic sequence
    :param restriction_enzyme: Name of restriction enzyme
                               (only when providing FASTA as regions)
    :param read_filters: List of :class:`~ReadFilter` to filter reads
                         while loading from SAM/BAM
    :param output_file: Path to output file
    :param check_sorted: Double-check that input SAM files
                         are sorted if True (default)
    :param threads: Number of threads used for finding restriction
                    fragments for read pairs
    :param batch_size: Number of read pairs sent to each restriction
                       fragment worker
    :return: :class:`~ReadPairs`
    """
    regions = genome_regions(regions, restriction_enzyme=restriction_enzyme)

    sb = SamBamReadPairGenerator(sam1_file, sam2_file, check_sorted=check_sorted)
    for f in read_filters:
        sb.add_filter(f)

    pairs = ReadPairs(file_name=output_file, mode='w')

    if isinstance(regions, RegionBased):
        pairs.add_regions(regions.regions, preserve_attributes=False)
    else:
        pairs.add_regions(regions, preserve_attributes=False)
    pairs.add_read_pairs(sb, threads=threads, batch_size=batch_size)

    return pairs


def generate_pairs_split(sam1_file, sam2_file, regions,
                         restriction_enzyme=None,
                         output_file=None, read_filters=(), check_sorted=True,
                         threads=1, batch_size=1000000):
    """
    Generate Pairs object from SAM/BAM files.

    This is a convenience function that let's you create a
    Pairs object from SAM/BAM data in a single step.

    :param sam1_file: Path to a sorted SAM/BAM file (1st mate)
    :param sam2_file: Path to a sorted SAM/BAM file (2nd mate)
    :param regions: Path to file with restriction fragments (BED, GFF)
                    or FASTA with genomic sequence
    :param restriction_enzyme: Name of restriction enzyme
                               (only when providing FASTA as regions)
    :param read_filters: List of :class:`~ReadFilter` to filter reads
                         while loading from SAM/BAM
    :param output_file: Path to output file
    :param check_sorted: Double-check that input SAM files
                         are sorted if True (default)
    :param threads: Number of threads used for finding restriction
                    fragments for read pairs
    :param batch_size: Number of read pairs sent to each restriction
                       fragment worker
    :return: :class:`~ReadPairs`
    """
    regions = genome_regions(regions, restriction_enzyme=restriction_enzyme)

    pairs = ReadPairs(file_name=output_file, mode='w')

    logger.debug("Adding regions")
    if isinstance(regions, RegionBased):
        for i, region in enumerate(regions.regions(lazy=True)):
            if i % 100000 == 0:
                logger.debug("{} regions".format(i))
            pairs._add_region(region, preserve_attributes=False)
        pairs.flush()
    else:
        pairs.add_regions(regions, preserve_attributes=False)

    pairs.add_read_pairs_from_sam(sam1_file, sam2_file, threads=threads, batch_size=batch_size,
                                  read_filters=read_filters, check_sorted=check_sorted)

    return pairs


class Monitor(WorkerMonitor):
    """
    Class to monitor fragment info worker threads.
    """
    def __init__(self, value=0, manager=None):
        WorkerMonitor.__init__(self, value=value, manager=manager)
        self.generating_pairs_lock = manager.Lock()

        with self.generating_pairs_lock:
            self.generating_pairs = True

    def set_generating_pairs(self, value):
        """
        Set the pair generating status.
        """
        with self.generating_pairs_lock:
            self.generating_pairs = value

    def is_generating_pairs(self):
        """
        Check the pair generating status.
        """
        with self.generating_pairs_lock:
            return self.generating_pairs


def _split_sam_worker(sam_file1, sam_file2, input_queue, monitor, batch_size=10000000,
                      tmpdir=None, check_sorted=True):
    monitor.set_generating_pairs(True)
    try:
        if tmpdir is None:
            tmpdir = tempfile.mkdtemp()

        output_prefix = os.path.join(tmpdir, 'split_pairs')
        try:
            logger.debug("Splitting and pairing SAM files")
            for pairs_file, chunk_size, unmappable in split_sam_pairs(sam_file1, sam_file2, output_prefix,
                                                                      chunk_size=batch_size,
                                                                      check_sorted=check_sorted):
                logger.debug("Split pairs batch {}".format(pairs_file))
                input_queue.put([pairs_file, chunk_size, unmappable])
                monitor.increment()
        except ValueError as e:
            logger.error(e)
            input_queue.put(None)
    finally:
        monitor.set_generating_pairs(False)


def _load_paired_sam_worker(monitor, input_file_queue, output_file_queue, fi, fe,
                            partition_breaks, read_filters=None,
                            tmpdir=None, buffer_size=1000000):
    worker_uuid = uuid.uuid4()
    monitor.set_worker_busy(worker_uuid)
    logger.debug("Starting load SAM worker {}".format(worker_uuid))

    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()

    file_counter = 0
    cumulative_wait_time = 0
    while True:
        # wait for input
        monitor.set_worker_idle(worker_uuid)
        logger.debug("Worker {} waiting for input".format(worker_uuid))
        s = datetime.now()
        read_pairs_file, chunk_size, unmappable = input_file_queue.get(True)
        w = datetime.now() - s
        logger.debug("Worker {} wait time: {}".format(worker_uuid, w))
        cumulative_wait_time += w.total_seconds()

        s = datetime.now()
        monitor.set_worker_busy(worker_uuid)
        logger.debug('Worker {} received input!'.format(worker_uuid))

        output_file = os.path.join(tmpdir, 'fragment_info_{}_{}.txt'.format(worker_uuid, file_counter))
        logger.debug("Writing fragment info to output file {}".format(output_file))
        file_counter += 1

        lines = []
        skipped_counter = 0
        pair_generator = PairedSamBamReadPairGenerator(read_pairs_file)
        if read_filters is not None:
            for f in read_filters:
                pair_generator.add_filter(f)
        pair_generator._unmappable_count = unmappable

        for read1, read2 in pair_generator:
            chrom1, pos1, flag1 = read1.reference_name, read1.pos, read1.flag
            chrom2, pos2, flag2 = read2.reference_name, read2.pos, read2.flag
            chrom1 = chrom1.decode() if isinstance(chrom1, bytes) else chrom1
            chrom2 = chrom2.decode() if isinstance(chrom2, bytes) else chrom2

            try:
                pos_ix1 = bisect_right(fe[chrom1], pos1)
                pos_ix2 = bisect_right(fe[chrom2], pos2)

                f_ix1, f_chromosome_ix1, f_start1, f_end1 = fi[chrom1][pos_ix1]
                f_ix2, f_chromosome_ix2, f_start2, f_end2 = fi[chrom2][pos_ix2]

                p_ix1 = bisect_right(partition_breaks, f_ix1)
                p_ix2 = bisect_right(partition_breaks, f_ix2)

                r_strand1 = -1 if flag1 & 16 else 1
                r_strand2 = -1 if flag2 & 16 else 1

                if p_ix1 > p_ix2:
                    line = "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(
                        p_ix2, pos2, r_strand2, f_ix2, f_chromosome_ix2, f_start2, f_end2,
                        p_ix1, pos1, r_strand1, f_ix1, f_chromosome_ix1, f_start1, f_end1
                    )
                else:
                    line = "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(
                        p_ix1, pos1, r_strand1, f_ix1, f_chromosome_ix1, f_start1, f_end1,
                        p_ix2, pos2, r_strand2, f_ix2, f_chromosome_ix2, f_start2, f_end2
                    )
                lines.append(line)

                if len(lines) > buffer_size:
                    with open(output_file, 'a') as o:
                        for line in lines:
                            o.write(line)
                    lines = []
            except (KeyError, IndexError):
                skipped_counter += 1

        with open(output_file, 'a') as o:
            for line in lines:
                o.write(line)

        logger.debug("Done obtaining fragment info for {} in {}".format(read_pairs_file, output_file))
        output_file_queue.put((read_pairs_file, output_file, pair_generator.stats()))

        l = datetime.now() - s
        logger.debug("Worker {} load time: {}".format(worker_uuid, l))


def _fragment_info_worker(monitor, input_queue, output_queue, fi, fe):
    """
    Worker that finds the restriction fragment info for read pairs.

    Finds the restriction fragment each read maps to, and returns the
    coordinates of the read and fragment pairs for each read pair.

    :param monitor: :class:`~Monitor`
    :param input_queue: Queue for input read_pairs
    :param output_queue: Queue for output fragment infos
    :param fi: Fragment info dictionary by chromosome and fragment index
    :param fe: Fragment end coordinates by chromosome
    :return: list of fragment infos
    """
    worker_uuid = uuid.uuid4()
    logger.debug("Starting fragment info worker {}".format(worker_uuid))

    while True:
        # wait for input
        monitor.set_worker_idle(worker_uuid)
        logger.debug("Worker {} waiting for input".format(worker_uuid))
        read_pairs = input_queue.get(True)
        monitor.set_worker_busy(worker_uuid)
        logger.debug('Worker {} reveived input!'.format(worker_uuid))
        read_pairs = msgpack.loads(read_pairs, strict_map_key=False)

        fragment_infos = []
        skipped_counter = 0
        for (chrom1, pos1, flag1), (chrom2, pos2, flag2) in read_pairs:
            chrom1 = chrom1.decode() if isinstance(chrom1, bytes) else chrom1
            chrom2 = chrom2.decode() if isinstance(chrom2, bytes) else chrom2

            try:
                pos_ix1 = bisect_right(fe[chrom1], pos1)
                pos_ix2 = bisect_right(fe[chrom2], pos2)

                f_ix1, f_chromosome_ix1, f_start1, f_end1 = fi[chrom1][pos_ix1]
                f_ix2, f_chromosome_ix2, f_start2, f_end2 = fi[chrom2][pos_ix2]

                r_strand1 = -1 if flag1 & 16 else 1
                r_strand2 = -1 if flag2 & 16 else 1

                fragment_infos.append(
                    ((pos1, r_strand1, f_ix1, f_chromosome_ix1, f_start1, f_end1),
                     (pos2, r_strand2, f_ix2, f_chromosome_ix2, f_start2, f_end2))
                )
            except (KeyError, IndexError):
                skipped_counter += 1
        logger.debug("Worker {} skipped {} pairs".format(worker_uuid, skipped_counter))
        output_queue.put(msgpack.dumps(fragment_infos))
        del read_pairs


def _read_pairs_worker(read_pairs, input_queue, monitor, batch_size=100000):
    """
    Worker to distribute incoming read pairs to fragment info workers.

    :param read_pairs: Iterator of read tuples (read1, read2)
    :param input_queue: Input queue for read pairs
    :param monitor: :class:`~Monitor`
    :param batch_size: Number of read pairs sent to each worker
    """
    logger.debug("Starting read pairs worker")
    try:
        read_pairs_batch = []
        for read1, read2 in read_pairs:
            read_pairs_batch.append((
                (read1.reference_name, read1.pos, read1.flag),
                (read2.reference_name, read2.pos, read2.flag)
            ))
            if len(read_pairs_batch) >= batch_size:
                logger.debug("Submitting read pair batch ({}) to input queue".format(batch_size))
                input_queue.put(msgpack.dumps(read_pairs_batch))
                read_pairs_batch = []
                monitor.increment()
        if len(read_pairs_batch) > 0:
            logger.debug("Submitting read pair batch ({}) to input queue".format(batch_size))
            input_queue.put(msgpack.dumps(read_pairs_batch))
            monitor.increment()
    finally:
        monitor.set_generating_pairs(False)
    logger.debug("Terminating read pairs worker")


class MinimalRead(object):
    """
    Minimal class representing an aligned read.

    .. attribute:: chromosome

        Chromosome name

    .. attribute:: reference_name

        Identical to chromosome, exists for
        compatibility with pysam

    .. attribute:: position

        Position of aligned read on chromosome

    .. attribute:: pos

        Identical to position, exists for
        compatibility with Pysam

    .. attribute:: strand

        Strand of the alignment

    .. attribute:: flag

        Flag representing the strandedness of a read.
        0 if on forward strand, else -1
    """

    def __init__(self, chromosome, position, strand):
        self.chromosome = chromosome
        self.reference_name = chromosome
        self.position = position
        self.strand = strand
        self.flag = 0 if strand == '+' or strand == 1 else -1
        self.pos = position
        self.tags = {}

    def get_tag(self, tag):
        raise KeyError("tag '{}' not present".format(tag))


class ReadPairGenerator(object):
    """
    Base class for generating and filtering read pairs.

    This class primarily provides filtering capabilities for
    read pairs generated by subclasses of :class:`~ReadPairGenerator`.
    You can add a :class:`~ReadFilter` using :func:`~ReadPairGenerator.add_filter`,
    which will be used during read pair generation. Filtering statistics
    are collected during the run, and can be obtained via
    :func:`~ReadPairGenerator.stats`.

    These generators are primarily meant as input for
    :func:`~ReadPairs.add_read_pairs`, but can also be used
    independently.

    Subclasses of :class:`~ReadPairGenerator` must implement the
    :func:`~ReadPairGenerator._iter_read_pairs` function.
    """
    def __init__(self):
        self.filters = []
        self._filter_stats = defaultdict(int)
        self._total_pairs = 0
        self._valid_pairs = 0

        self._unmappable_count = 0

    def _iter_read_pairs(self, *args, **kwargs):
        raise NotImplementedError("Class must override iter_read_pairs")

    def add_filter(self, read_filter):
        """
        Add a :class:`~ReadFilter` to this object.

        The filter will be applied during read pair generation.

        :param read_filter: :class:`~ReadFilter`
        """
        if not isinstance(read_filter, ReadFilter):
            raise ValueError("argument must be an instance of "
                             "class ReadFilter, not {}!".format(type(read_filter)))
        self.filters.append(read_filter)

    def stats(self):
        """
        Return filter statistics collected during read pair generation.

        The :func:`~ReadPairGenerator.__iter__` filters reads based on the
        filters added using :func:`~ReadPairGenerator.add_filter`. During
        filtering, it keeps track of the numbers of read pairs that were
        filtered. This function returns a :class:`~dict` of the form
        <filter_name>: <number of reads filtered out>.

        :return: dict
        """
        filter_names = []
        for i, f in enumerate(self.filters):
            if hasattr(f, 'mask') and f.mask is not None:
                filter_names.append(f.mask.name)
            else:
                filter_names.append('filter_{}'.format(i))

        stats = dict()
        for i, count in self._filter_stats.items():
            stats[filter_names[i]] = count
        stats['unmappable'] = self._unmappable_count
        stats['valid'] = self._valid_pairs
        stats['total'] = self._total_pairs + self._unmappable_count
        return stats

    def __iter__(self):
        filters = []
        for f in self.filters:
            if not isinstance(f, UnmappedFilter):
                filters.append(f)
            else:
                logger.warning("Unmappable reads are filtered by default, "
                               "no need to add explicit filter.")
        self.filters = filters

        self._filter_stats = defaultdict(int)
        self._total_pairs = 0
        self._valid_pairs = 0
        for (read1, read2) in self._iter_read_pairs():
            valid_reads = True
            for i, f in enumerate(self.filters):
                if not f.valid_read(read1) or not f.valid_read(read2):
                    self._filter_stats[i] += 1
                    valid_reads = False
            if valid_reads:
                yield (read1, read2)
                self._valid_pairs += 1
            self._total_pairs += 1


class TxtReadPairGenerator(ReadPairGenerator):
    """
    Generate read pairs from a plain text file.

    This is an implementation of :class:`~ReadPairGenerator` for
    reading read pairs from an arbitrary text file. For specific
    text file formats have a look at :class:`~FourDNucleomePairGenerator`
    or :class:`~HicProPairGenerator`.

    :class:`~TxtReadPairGenerator` iterates over lines in
    a file and splits them into fields at "sep". It then extracts
    the chromosome, position and strand for each read in the pair,
    according to the fields specified by "chr<1|2>_field",
    "pos<1|2>_field", and "strand<1|2>_field". If your file does not
    have strand fields, or if you don't want to load them, you can
    simply set them to "None".
    """
    def __init__(self, valid_pairs_file, sep=None,
                 chr1_field=1, pos1_field=2, strand1_field=3,
                 chr2_field=4, pos2_field=5, strand2_field=6):
        """
        Initialise read pair generator.

        :param valid_pairs_file: Path to a txt file with valid pairs.
                                 May be gzipped, in which case the file
                                 name must end with ".gz" or ".gzip"
        :param sep: A field separator. Defaults to whitespace.
        :param chr1_field: int, index of the field which has the information
                           for the first mate's chromosome
        :param pos1_field: int, index of the field which has the information
                           for the first mate's position on the reference in bp
        :param strand1_field: int, index of the field which has the information
                              for the first mate's strand (+/-)
        :param chr2_field: int, index of the field which has the information
                           for the second mate's chromosome
        :param pos2_field: int, index of the field which has the information
                           for the second mate's position on the reference in bp
        :param strand2_field: int, index of the field which has the information
                              for the second mate's strand (+/-)
        """
        ReadPairGenerator.__init__(self)
        self._file_name = valid_pairs_file
        self.sep = sep
        self.chr1_field = chr1_field
        self.pos1_field = pos1_field
        self.strand1_field = strand1_field
        self.chr2_field = chr2_field
        self.pos2_field = pos2_field
        self.strand2_field = strand2_field

        if self._file_name.endswith('.gz') or self._file_name.endswith('gzip'):
            self._open_file = gzip.open
        else:
            self._open_file = open

        if not os.path.exists(valid_pairs_file):
            raise ValueError("File {} does not exist!".format(valid_pairs_file))

        # check that this file is valid:
        with self._open_file(valid_pairs_file, 'rt') as f:
            for line in f:
                line = line.rstrip()
                if line.startswith('#') or line == '':
                    continue

                fields = line.split(sep)

                # find max field
                max_field_ix = 0
                for field_number in {chr1_field, chr2_field,
                                     pos1_field, pos2_field,
                                     strand1_field, strand2_field}:
                    if field_number is not None:
                        max_field_ix = max(max_field_ix, field_number)

                if len(fields) < max_field_ix + 1:
                    raise ValueError("Not enough fields ({}) in file {}".format(len(fields),
                                                                                valid_pairs_file))

                # ensure we can transform pos fields to int
                int(fields[pos1_field])
                int(fields[pos2_field])
                # ensure strand fields are valid if specified
                if strand1_field is not None and fields[strand1_field] not in {'+', '-', '.', '-1', '1', '+1'}:
                    raise ValueError("Cannot read strand1 field!")
                if strand2_field is not None and fields[strand2_field] not in {'+', '-', '.', '-1', '1', '+1'}:
                    raise ValueError("Cannot read strand2 field!")

                break

    def _iter_read_pairs(self, *args, **kwargs):
        """
        Iterate over read pairs encoded in each line of the txt file.

        :return: read pair iterator over tuples of the form
                 (:class:`~MinimalRead`, :class:`~MinimalRead`)
        """
        with self._open_file(self._file_name, 'rt') as f:
            for line in f:
                line = line.rstrip()
                if line == '' or line.startswith('#'):
                    continue
                fields = line.split(self.sep)

                strand1 = fields[self.strand1_field] if self.strand1_field is not None else '+'
                strand2 = fields[self.strand2_field] if self.strand2_field is not None else '+'

                read1 = MinimalRead(chromosome=fields[self.chr1_field],
                                    position=int(fields[self.pos1_field]),
                                    strand=strand1)
                read2 = MinimalRead(chromosome=fields[self.chr2_field],
                                    position=int(fields[self.pos2_field]),
                                    strand=strand2)
                yield (read1, read2)


class HicProPairGenerator(TxtReadPairGenerator):
    """
    Read pair generator for HiC-Pro "validPairs" files.

    This generator is a subclass of :class:`~TxtReadPairGenerator`
    with presets for fields in HiC-Pro validPairs files.
    """
    def __init__(self, file_name):
        """
        Inititalise this HiC-Pro read pair generator.

        :param file_name: Path to HiC-Pro ".validPairs" file
        """
        TxtReadPairGenerator.__init__(self, file_name, sep="\t",
                                      chr1_field=1, pos1_field=2, strand1_field=3,
                                      chr2_field=4, pos2_field=5, strand2_field=6)


class FourDNucleomePairGenerator(TxtReadPairGenerator):
    """
    Read pair generator that works on 4D Nucleome ".pairs" files.

    For details on the 4D Nucleome pairs format see:
    https://github.com/4dn-dcic/pairix/blob/master/pairs_format_specification.md

    """
    def __init__(self, pairs_file):
        if pairs_file.endswith('.gz') or pairs_file.endswith('gzip'):
            open_file = gzip.open
        else:
            open_file = open

        columns = dict()
        with open_file(pairs_file, 'rt') as f:
            columns_line = None
            for line_ix, line in enumerate(f):
                if line_ix == 0 and not line.startswith("## pairs format"):
                    raise ValueError("Not a 4D nucleome pairs format file."
                                     "Missing '## pairs format X.X' header line.")

                if line.startswith('#columns:'):
                    columns_line = line

            if columns_line is None:
                raise ValueError("Pairs file does not contain a "
                                 "'#columns' entry in the header")

            _, columns_field = columns_line.split(':')
            for i, name in enumerate(columns_field.split()):
                columns[name] = i

        chromosome1_field_id = 'chr1' if 'chr1' in columns else 'chrom1'
        chromosome2_field_id = 'chr2' if 'chr2' in columns else 'chrom2'

        TxtReadPairGenerator.__init__(self, pairs_file, sep=None,
                                      chr1_field=columns[chromosome1_field_id],
                                      pos1_field=columns['pos1'],
                                      strand1_field=columns['strand1'] if 'strand1' in columns else None,
                                      chr2_field=columns[chromosome2_field_id],
                                      pos2_field=columns['pos2'],
                                      strand2_field=columns['strand2'] if 'strand2' in columns else None,
                                      )


class PairedSamBamReadPairGenerator(ReadPairGenerator):
    """
    Generate read pairs from single SAM/BAM files from reads with identical qname.

    Chimeric reads (mapping partially to multiple genomic locations),
    such as output by BWA, are handled as follows: If both mate pairs
    are chimeric, they are removed. If only one mate is chimeric, but
    it is split into more than 2 alignments, it is also removed. If
    one mate is chimeric and split into two alignments. If one part
    of the chimeric alignment maps within 100bp of the regular alignment,
    the read pair is kept and returned. In all other cases, the pair is
    removed.
    """
    def __init__(self, sam_file):
        ReadPairGenerator.__init__(self)
        self.sam_file = sam_file

    @staticmethod
    def resolve_chimeric(reads, max_dist_same_locus=500):
        """
        :return: read1, read2, is_chimeric
        """
        if len(reads) == 2:
            return reads[0], reads[1], False
        elif len(reads) == 3:
            if reads[0].reference_id != reads[1].reference_id != reads[2].reference_id:
                return None, None, False
            else:
                start0, stop0 = reads[0].pos, reads[0].pos + reads[0].alen
                start1, stop1 = reads[1].pos, reads[1].pos + reads[1].alen
                start2, stop2 = reads[2].pos, reads[2].pos + reads[2].alen

                d0_1 = min(abs(start0 - stop1), abs(start1 - stop0))
                d0_2 = min(abs(start0 - stop2), abs(start2 - stop0))
                d1_2 = min(abs(start1 - stop2), abs(start2 - stop1))

                # inter-chromosomal or complex
                if reads[0].reference_id == reads[1].reference_id != reads[2].reference_id:
                    if d0_1 < max_dist_same_locus:
                        return reads[0], reads[2], True
                    else:
                        return None, None, False
                elif reads[0].reference_id != reads[1].reference_id == reads[2].reference_id:
                    if d1_2 < max_dist_same_locus:
                        return reads[0], reads[1], True
                    else:
                        return None, None, False
                elif reads[0].reference_id == reads[2].reference_id != reads[1].reference_id:
                    if d0_2 < max_dist_same_locus:
                        return reads[1], reads[2], True
                    else:
                        return None, None, False
                # intra-chromosomal
                else:
                    if d0_1 <= max_dist_same_locus and d0_2 > max_dist_same_locus and d1_2 > max_dist_same_locus:
                        return reads[0], reads[2], True
                    elif d0_2 <= max_dist_same_locus  and d0_1 > max_dist_same_locus and d1_2 > max_dist_same_locus:
                        return reads[1], reads[2], True
                    elif d1_2 <= max_dist_same_locus  and d0_1 > max_dist_same_locus and d0_2 > max_dist_same_locus:
                        return reads[0], reads[1], True
                    else:
                        return None, None, None
        else:
            return None, None, False

    def _iter_read_pairs(self, *args, **kwargs):
        if isinstance(self.sam_file, pysam.AlignmentFile):
            sam = self.sam_file
        else:
            sam = pysam.AlignmentFile(self.sam_file)

        chimeric_pairs = 0
        normal_pairs = 0
        abnormal_pairs = 0

        current_qname = None
        reads = []
        for read in sam:
            qname = read.qname.encode('utf-8')
            if current_qname is None or natural_cmp(qname, current_qname) == 0:
                reads.append(read)
            else:
                if len(reads) >= 2:
                    read1, read2, is_chimeric = PairedSamBamReadPairGenerator.resolve_chimeric(reads)
                    if read1 is not None and read2 is not None:
                        yield (read1, read2)
                        if is_chimeric:
                            chimeric_pairs += 1
                        else:
                            normal_pairs += 1
                    else:
                        abnormal_pairs += 1
                reads = [read]
            current_qname = qname

        # final reads
        if len(reads) >= 2:
            read1, read2, is_chimeric = PairedSamBamReadPairGenerator.resolve_chimeric(reads)
            if read1 is not None and read2 is not None:
                yield (read1, read2)
                if is_chimeric:
                    chimeric_pairs += 1
                else:
                    normal_pairs += 1
            else:
                abnormal_pairs += 1

        logger.debug("Done generating read pairs.")
        logger.debug("Normal pairs: {}".format(normal_pairs))
        logger.debug("Chimeric pairs: {}".format(chimeric_pairs))
        logger.debug("Abnormal pairs: {}".format(abnormal_pairs))


class SamBamReadPairGenerator(ReadPairGenerator):
    """
    Generate read pairs from paired-end SAM/BAM files.

    This :class:`~ReadPairGenerator` iterates over two SAM or BAM
    files that have been sorted by qname (for example with
    :code:`samtools sort -n`).

    Chimeric reads (mapping partially to multiple genomic locations),
    such as output by BWA, are handled as follows: If both mate pairs
    are chimeric, they are removed. If only one mate is chimeric, but
    it is split into more than 2 alignments, it is also removed. If
    one mate is chimeric and split into two alignments. If one part
    of the chimeric alignment maps within 100bp of the regular alignment,
    the read pair is kept and returned. In all other cases, the pair is
    removed.
    """
    def __init__(self, sam_file1, sam_file2, check_sorted=True):
        ReadPairGenerator.__init__(self)
        self.sam_file1 = sam_file1
        self.sam_file2 = sam_file2
        self._check_sorted = check_sorted
        if not os.path.exists(self.sam_file1):
            raise ValueError("File {} does not exist!".format(self.sam_file1))
        if not os.path.exists(self.sam_file2):
            raise ValueError("File {} does not exist!".format(self.sam_file2))

    def _iter_read_pairs(self, *args, **kwargs):
        max_dist_same_locus = kwargs.get('max_dist_same_locus', 100)
        logger.info("Starting to generate read pairs from SAM")

        def _all_reads(iterator, last_read=None):
            reads = []
            if last_read is not None:
                if last_read.is_unmapped:
                    self._unmappable_count += 1
                else:
                    reads.append(last_read)

            next_read = None
            try:
                next_read = next(iterator)
                while len(reads) == 0 or natural_cmp(next_read.qname.encode(), reads[0].qname.encode()) == 0:
                    if not next_read.is_unmapped:
                        reads.append(next_read)
                    else:
                        self._unmappable_count += 1
                    next_read = next(iterator)
            except StopIteration:
                if len(reads) == 0:
                    raise
            return reads[0].qname.encode(), reads, next_read

        def _find_pair(reads1, reads2):
            """
            :return: read1, read2, is_chimeric
            """
            if len(reads1) == len(reads2) == 1:
                return reads1[0], reads2[0], False
            elif (len(reads1) == 1 and len(reads2) == 2) or (len(reads2) == 1 and len(reads1) == 2):
                if len(reads2) > len(reads1):
                    reads1, reads2 = reads2, reads1

                read2 = reads2[0]
                match_pos2 = find_alignment_match_positions(read2, longest=True)[0]
                if match_pos2 is None:
                    return None, None, True

                read1 = None
                same_locus = False
                for read in reads1:
                    if read.reference_id != read2.reference_id:
                        read1 = read
                    else:
                        match_pos1 = find_alignment_match_positions(read, longest=True)[0]
                        if match_pos1 is None:
                            return None, None, True

                        if min(abs(match_pos1[0] - match_pos2[1]),
                               abs(match_pos1[1] - match_pos2[0])) > max_dist_same_locus:
                            read1 = read
                        else:
                            same_locus = True

                if same_locus:
                    return read1, read2, True
            return None, None, False

        with pysam.AlignmentFile(self.sam_file1) as sam1:
            with pysam.AlignmentFile(self.sam_file2) as sam2:
                normal_pairs = 0
                chimeric_pairs = 0
                abnormal_pairs = 0

                sam1_iter = iter(sam1)
                sam2_iter = iter(sam2)
                try:
                    qname1, reads1, next_read1 = _all_reads(sam1_iter)
                    qname2, reads2, next_read2 = _all_reads(sam2_iter)
                    while True:
                        check1 = False
                        check2 = False
                        previous_qname1 = qname1
                        previous_qname2 = qname2

                        cmp = natural_cmp(qname1, qname2)
                        if cmp == 0:  # read name identical
                            read1, read2, is_chimeric = _find_pair(reads1, reads2)
                            if read1 is not None and read2 is not None:
                                yield (read1, read2)
                                if is_chimeric:
                                    chimeric_pairs += 1
                                else:
                                    normal_pairs += 1
                            else:
                                abnormal_pairs += 1
                            qname1, reads1, next_read1 = _all_reads(sam1_iter, last_read=next_read1)
                            qname2, reads2, next_read2 = _all_reads(sam2_iter, last_read=next_read2)
                            check1, check2 = True, True
                        elif cmp < 0:  # first pointer behind
                            qname1, reads1, next_read1 = _all_reads(sam1_iter, last_read=next_read1)
                            check1 = True
                        else:  # second pointer behind
                            qname2, reads2, next_read2 = _all_reads(sam2_iter, last_read=next_read2)
                            check2 = True

                        # check that the files are sorted
                        if self._check_sorted:
                            if check1 and natural_cmp(previous_qname1, qname1) > 0:
                                raise ValueError("First SAM file is not sorted by "
                                                 "read name (samtools sort -n)! Read names:"
                                                 "{} and {}".format(previous_qname1, qname1))
                            if check2 and natural_cmp(previous_qname2, qname2) > 0:
                                raise ValueError("Second SAM file is not sorted by "
                                                 "read name (samtools sort -n)! Read names:"
                                                 "{} and {}".format(previous_qname2, qname2))
                except StopIteration:
                    logger.info("Done generating read pairs.")
                    logger.info("Normal pairs: {}".format(normal_pairs))
                    logger.info("Chimeric pairs: {}".format(chimeric_pairs))
                    logger.info("Abnormal pairs: {}".format(abnormal_pairs))

    def filter_quality(self, cutoff=30):
        """
        Convenience function that registers a QualityFilter.
        The actual algorithm and rationale used for filtering will depend on the
        internal _mapper attribute.

        :param cutoff: Minimum mapping quality (mapq) a read must have to pass
                       the filter
        """
        mask = Mask('map quality', 'Mask read pairs with a mapping quality lower than {}'.format(cutoff),
                    ix=len(self.filters))
        quality_filter = QualityFilter(cutoff, mask)
        self.add_filter(quality_filter)

    def filter_unmapped(self):
        """
        Convenience function that registers an UnmappedFilter.
        """
        mask = Mask('unmappable', 'Mask read pairs that are unmapped', ix=len(self.filters))
        unmapped_filter = UnmappedFilter(mask)
        self.add_filter(unmapped_filter)

    def filter_multi_mapping(self, strict=True):
        """
        Convenience function that registers a UniquenessFilter.
        The actual algorithm and rationale used for filtering will depend on the
        internal _mapper attribute.

        :param strict: If True will filter if XS tag is present. If False,
                       will filter only when XS tag is not 0. This is applied if
                       alignments are from bowtie2.
        """
        mask = Mask('multi-mapping', 'Mask reads that do not map uniquely (according to XS tag)',
                    ix=len(self.filters))
        uniqueness_filter = UniquenessFilter(strict, mask)
        self.add_filter(uniqueness_filter)


class FragmentRead(object):
    """
    Class representing a fragment-mapped read.

    .. attribute:: fragment

        A :class:`~fanc.GenomicRegion` delineated by
        restriction sites.

    .. attribute:: position

        The position of this read in base-pairs (1-based) from the
        start of the chromosome it maps to.

    .. attribute:: strand

        The strand this read maps to (-1 or +1).

    """

    def __init__(self, fragment=None, position=None, strand=0):
        """
        Initialize this :class:`~FragmentRead` object.

        :param fragment: A :class:`~fanc.GenomicRegion` delineated by
                         restriction sites.
        :param position: The position of this read in base-pairs (1-based) from the
                         start of the chromosome it maps to.
        :param strand: The strand this read maps to (-1 or +1).
        """
        self.fragment = fragment
        self.position = position
        self.strand = strand

    def re_distance(self):
        """
        Get the distance of the alignment to the nearest restriction site.
        :return: int
        """
        return min(abs(self.position - self.fragment.start),
                   abs(self.position - self.fragment.end))

    def __repr__(self):
        return "%s: %d-(%d[%d])-%d" % (self.fragment.chromosome,
                                       self.fragment.start,
                                       self.position,
                                       self.strand,
                                       self.fragment.end)


class LazyFragmentRead(FragmentRead):
    """
    :class:`~FragmentRead` implementation with lazy attribute loading.

    .. attribute:: fragment

        A :class:`~fanc.GenomicRegion` delineated by
        restriction sites.

    .. attribute:: position

        The position of this read in base-pairs (1-based) from the
        start of the chromosome it maps to.

    .. attribute:: strand

        The strand this read maps to (-1 or +1).

    """
    def __init__(self, row, pairs, side="left"):
        self._row = row
        self._pairs = pairs
        self._side = side

    @property
    def position(self):
        return self._row[self._side + "_read_position"]

    @property
    def strand(self):
        return self._row[self._side + "_read_strand"]

    @property
    def fragment(self):
        return LazyFragment(self._row, self._pairs, side=self._side)


class LazyFragment(GenomicRegion):
    """
    :class:`~fanc.GenomicRegion` representing a fragment with lazy attribute loading.

    .. attribute:: chromosome

        The reference sequence this region is located on

    .. attribute:: start

        start position of this region on the reference (1-based, inclusive)

    .. attribute:: end

        end position of this region on the reference (1-based, inclusive)

    .. attribute:: strand

        strand of the reference this region is located on (-1, 1, 0, or None)

    .. attribute:: ix

        Region index within the context of regions from the same object.
    """
    def __init__(self, row, pairs, ix=None, side="left"):
        self._row = row
        self._pairs = pairs
        self._side = side
        self._static_ix = ix

    @property
    def chromosome(self):
        return self._pairs._ix_to_chromosome[self._row[self._side + "_fragment_chromosome"]]

    @property
    def start(self):
        return self._row[self._side + "_fragment_start"]

    @property
    def end(self):
        return self._row[self._side + "_fragment_end"]

    @property
    def strand(self):
        return 1

    @property
    def ix(self):
        if self._static_ix is None:
            return self._row['source'] if self._side == 'left' else self._row['sink']
        return self._static_ix


class ReadPairs(RegionPairsTable):
    """
    Class representing a collection of read pairs mapped to restriction fragments.

    This class is a :class:`~fanc.RegionBased` object, where each
    :class:`~fanc.GenomicRegion` represents a restriction fragment
    from a Hi-C experiment. A list of fragments can be obtained with
    the :func:`~fanc.regions.genome_regions` function, for example.

    To create a :class:`~ReadPairs` object, you first have to add
    the restriction fragments before adding read pairs:

    .. code::

        import fanc

        re_fragments = fanc.genome_regions("hg19_chr18_19.fa", "HindIII")

        rp = fanc.ReadPairs()
        rp.add_regions(re_fragments.regions)

    Read pairs can easily be generate fro different types of input using
    :class:`~ReadPairGenerator` implementations, e.g.
    :class:`~HicProReadPairGenerator` or :class:`~SamBamReadPairGenerator`.

    .. code::

        rp_generator = fanc.SamBamReadPairGenerator("output/sam/SRR4271982_chr18_19_1_sort.bam",
                                                    "output/sam/SRR4271982_chr18_19_2_sort.bam")
        rp.add_read_pairs(rp_generator, threads=4)


    You can query regions using the :class:`~fanc.RegionBased` interface:

    .. code::

        chr1_fragments = rp.regions("chr1")

    and you can iterate over read pairs using the :func:`~ReadPairs.pairs`:

    .. code:

        for pair in rp.pairs(lazy=True):
            print(pair)

    you can also use the :cls:Region
    """
    _classid = 'READPAIRS'

    def __init__(self, file_name=None, mode='a',
                 _group_name='fragment_map',
                 _table_name_fragments='fragments',
                 _table_name_pairs='pairs',
                 tmpdir=None):
        """
        Initialize empty FragmentMappedReadPairs object.

        :param file_name: Path to a file that will be created to save
                          this object or path to an existing HDF5 file
                          representing a FragmentMappedReadPairs object.
        :param mode: File mode. Defaults to 'a' (append). Use 'w' to overwrite
                     an existing file in the same location, and 'r' for safe
                     read-only access.
        """
        RegionPairsTable.__init__(self, file_name=file_name, mode=mode, tmpdir=tmpdir,
                                  additional_edge_fields={
                                      'ix': t.Int32Col(pos=0),
                                      'left_read_position': t.Int64Col(pos=1),
                                      'left_read_strand': t.Int8Col(pos=2),
                                      'left_fragment_start': t.Int64Col(pos=3),
                                      'left_fragment_end': t.Int64Col(pos=4),
                                      'left_fragment_chromosome': t.Int32Col(pos=5),
                                      'right_read_position': t.Int64Col(pos=6),
                                      'right_read_strand': t.Int8Col(pos=7),
                                      'right_fragment_start': t.Int64Col(pos=8),
                                      'right_fragment_end': t.Int64Col(pos=9),
                                      'right_fragment_chromosome': t.Int32Col(pos=10)
                                  },
                                  _table_name_regions=_table_name_fragments,
                                  _table_name_edges=_table_name_pairs)

        self._pairs = self._edges
        if self._partition_breaks is None:
            self._pair_count = 0
        else:
            self._pair_count = None

        self._ix_to_chromosome = dict()
        self._chromosome_to_ix = dict()
        self._update_references()

    def _update_references(self):
        """
        Update internal chromosome index dictionaries.
        """
        if self._chromosomes_info is not None:
            for row in self._chromosomes_info.iterrows():
                ix, chromosome = row['ix'], row['name'].decode()
                self._ix_to_chromosome[ix] = chromosome
                self._chromosome_to_ix[chromosome] = ix
        else:
            chromosomes = []
            for region in self.regions(lazy=True):
                if len(chromosomes) == 0 or chromosomes[-1] != region.chromosome:
                    chromosomes.append(region.chromosome)

            for i, chromosome in enumerate(chromosomes):
                self._ix_to_chromosome[i] = chromosome
                self._chromosome_to_ix[chromosome] = i

    def _flush_regions(self):
        """
        Write buffered regions to file and update region references.
        """
        if self._regions_dirty:
            RegionPairsTable._flush_regions(self)
            self._update_references()

    def flush(self, silent=config.hide_progressbars):
        """
        Write buffered data to file and update indexes,

        :param silent: If True, does not use progressbars.
        """
        RegionPairsTable.flush(self, silent=silent)

    def _read_fragment_info(self, read):
        chromosome = read.reference_name
        fragment_info = None
        for row in self._regions.where(
                        "(start <= %d) & (end >= %d) & (chromosome == b'%s')" % (read.pos, read.pos, chromosome)):
            fragment_info = [row['ix'], self._chromosome_to_ix[chromosome], row['start'], row['end']]

        if fragment_info is None:
            raise ValueError("No matching region can be found for {}".format(read))

        return fragment_info

    def _read_pair_fragment_info(self, read_pair):
        read1, read2 = read_pair
        f_ix1, f_chromosome_ix1, f_start1, f_end1 = self._read_fragment_info(read1)
        f_ix2, f_chromosome_ix2, f_start2, f_end2 = self._read_fragment_info(read2)
        r_strand1 = -1 if read1.flag & 16 else 1
        r_strand2 = -1 if read2.flag & 16 else 1
        return ((read1.pos, r_strand1, f_ix1, f_chromosome_ix1, f_start1, f_end1),
                (read2.pos, r_strand2, f_ix2, f_chromosome_ix2, f_start2, f_end2))

    def _read_pairs_fragment_info(self, read_pairs, threads=4, batch_size=1000000, timeout=600):
        """
        Parallel loading of read pairs along with mapping to restriction fragments.

        :param read_pairs: iterator of read pairs, typically
                           from a :class:`~ReadPairGenerator`
        :param threads: Number of threads used for parallel
                        fragment info finding.
        :param batch_size: Number of read pairs sent to each worker
        :param timeout: Time to wait for reply of first worker. If this
                        threshold is exceeded before any read pairs have been
                        returned, a warning is displayed.
        """
        fragment_infos = defaultdict(list)
        fragment_ends = defaultdict(list)
        for region in self.regions(lazy=True):
            chromosome = region.chromosome
            fragment_infos[chromosome].append((region.ix, self._chromosome_to_ix[chromosome],
                                               region.start, region.end))
            fragment_ends[chromosome].append(region.end)

        worker_pool = None
        t_pairs = None
        try:
            monitor = Monitor()
            input_queue = mp.Queue(maxsize=2*threads)
            output_queue = mp.Queue(maxsize=2*threads)

            monitor.set_generating_pairs(True)
            t_pairs = threading.Thread(target=_read_pairs_worker, args=(read_pairs, input_queue,
                                                                        monitor, batch_size))
            t_pairs.daemon = True
            t_pairs.start()

            worker_pool = mp.Pool(threads, _fragment_info_worker,
                                  (monitor, input_queue, output_queue, fragment_infos, fragment_ends))

            output_counter = 0
            while output_counter < monitor.value() or not monitor.workers_idle() or monitor.is_generating_pairs():
                try:
                    read_pair_infos = output_queue.get(block=True, timeout=timeout)

                    for read1_info, read2_info in msgpack.loads(read_pair_infos, strict_map_key=False):
                        yield read1_info, read2_info
                    output_counter += 1
                    del read_pair_infos
                except Empty:
                    logger.warning("Reached SAM pair generator timeout. This could mean that no "
                                   "valid read pairs were found after filtering. "
                                   "Check filter settings!")
        finally:
            if worker_pool is not None:
                worker_pool.terminate()
            if t_pairs is not None:
                t_pairs.join()

    def _add_infos(self, fi1, fi2):
        r_pos1, r_strand1, f_ix1, f_chromosome_ix1, f_start1, f_end1 = fi1
        r_pos2, r_strand2, f_ix2, f_chromosome_ix2, f_start2, f_end2 = fi2

        if self._pair_count is None:
            self._pair_count = sum(edge_table._original_len()
                                   for _, edge_table in self._iter_edge_tables())

        edge = Edge(ix=self._pair_count,
                    source=f_ix1, sink=f_ix2,
                    left_read_position=r_pos1, right_read_position=r_pos2,
                    left_read_strand=r_strand1, right_read_strand=r_strand2,
                    left_fragment_start=f_start1, right_fragment_start=f_start2,
                    left_fragment_end=f_end1, right_fragment_end=f_end2,
                    left_fragment_chromosome=f_chromosome_ix1,
                    right_fragment_chromosome=f_chromosome_ix2)

        self._add_pair(edge)

    def _default_edge_list(self):
        record = [None] * len(self._field_names_dict)
        for name, ix in self._field_names_dict.items():
            record[ix] = self._edge_field_defaults[name]
        return record

    def add_read_pair(self, read_pair, flush=True):
        fi1, fi2 = self._read_pair_fragment_info(read_pair)
        self._add_infos(fi1, fi2)
        if flush:
            self.flush()

    def load_read_pairs_fragment_info_file(self, read_pairs_file):
        if read_pairs_file.endswith('.gz') or read_pairs_file.endswith('.gzip'):
            open_ = gzip.open
        else:
            open_ = open

        if self._pair_count is None:
            self._pair_count = sum(edge_table._original_len()
                                   for _, edge_table in self._iter_edge_tables())

        edge = {}
        with open_(read_pairs_file) as f:
            for line in f:
                line = line.rstrip()
                if line == '':
                    continue

                fields = line.split("\t")
                info1, info2 = fields[:7], fields[7:]
                if int(info1[3]) > int(info2[3]):
                    info1, info2 = info2, info1

                edge['ix'] = self._pair_count
                edge['source'] = int(info1[3])
                edge['sink'] = int(info2[3])
                edge['left_read_position'], edge['right_read_position'] = int(info1[1]), int(info2[1])
                edge['left_read_strand'], edge['right_read_strand'] = info1[2], info2[2]
                edge['left_fragment_start'], edge['right_fragment_start'] = int(info1[5]), int(info2[5])
                edge['left_fragment_end'], edge['right_fragment_end'] = int(info1[6]), int(info2[6])
                edge['left_fragment_chromosome'], edge['right_fragment_chromosome'] = int(info1[4]), int(info2[4])

                self._edge_buffer.add_dict(edge, partition=(int(info1[0]), int(info2[0])))
                self._pair_count += 1

    def add_read_pairs_from_sam(self, sam_file1, sam_file2, batch_size=1000000, threads=1,
                                read_filters=None, check_sorted=True, tmpdir=None):
        self._edges_dirty = True
        self._disable_edge_indexes()

        fragment_infos = defaultdict(list)
        fragment_ends = defaultdict(list)
        for region in self.regions(lazy=True):
            chromosome = region.chromosome
            fragment_infos[chromosome].append((region.ix, self._chromosome_to_ix[chromosome],
                                               region.start, region.end))
            fragment_ends[chromosome].append(region.end)

        if tmpdir is None:
            split_tmpdir = tempfile.mkdtemp()
            pairs_tmpdir = tempfile.mkdtemp()
        else:
            split_tmpdir = tempfile.mkdtemp(dir=tmpdir)
            pairs_tmpdir = tempfile.mkdtemp(dir=tmpdir)

        worker_pool = None
        t_split = None
        all_stats = defaultdict(int)
        try:
            queue_manager = mp.Manager()
            input_file_queue = queue_manager.Queue(maxsize=threads * 3)
            output_file_queue = queue_manager.Queue(maxsize=threads * 3)

            monitor = Monitor(manager=queue_manager)
            monitor.set_generating_pairs(True)
            t_split = threading.Thread(target=_split_sam_worker, args=(sam_file1, sam_file2,
                                                                       input_file_queue,
                                                                       monitor, batch_size,
                                                                       split_tmpdir,
                                                                       check_sorted))
            t_split.daemon = True
            t_split.start()

            worker_pool = mp.Pool(threads, _load_paired_sam_worker,
                                  (monitor, input_file_queue, output_file_queue,
                                   fragment_infos, fragment_ends, self._partition_breaks,
                                   read_filters, pairs_tmpdir))

            output_counter = 0
            cumulative_wait_time = 0
            cumulative_load_time = 0
            while output_counter < monitor.value() or not monitor.workers_idle() or monitor.is_generating_pairs():
                try:
                    s = datetime.now()
                    input_file, read_pairs_file, chunk_stats = output_file_queue.get(block=True)
                    w = datetime.now() - s
                    logger.debug("Wait time: {}".format(w))
                    cumulative_wait_time += w.total_seconds()

                    s = datetime.now()
                    os.remove(input_file)
                    self.load_read_pairs_fragment_info_file(read_pairs_file)
                    os.remove(read_pairs_file)
                    for key, value in chunk_stats.items():
                        all_stats[key] += value
                    output_counter += 1
                    l = datetime.now() - s
                    logger.debug("Load time: {}".format(l))
                    cumulative_load_time += l.total_seconds()
                except Empty:
                    logger.warning("Reached SAM pair generator timeout. This could mean that no "
                                   "valid read pairs were found after filtering. "
                                   "Check filter settings!")
        finally:
            if worker_pool is not None:
                worker_pool.terminate()
            if t_split is not None:
                t_split.join()
            shutil.rmtree(split_tmpdir)
            shutil.rmtree(pairs_tmpdir)

        logger.debug('Cumulative: {} wait, {} load'.format(cumulative_wait_time, cumulative_load_time))

        if 'read_filter_stats' not in self.meta:
            self.meta.read_filter_stats = all_stats
        else:
            self.meta.read_filter_stats = add_dict(self.meta.read_filter_stats, all_stats)

        self.flush()

    def add_read_pairs(self, read_pairs, batch_size=1000000, threads=1):
        """
        Add read pairs to this object.

        This function requires tuples of read pairs as input, for
        example from :class:`~MinimalRead` or :class:`~pysam.AlignedSegment`.
        Typically, you won't have to construct these from scratch, but can
        use one of the :class:`~ReadPairGenerator` classes to generate
        read pairs from input files.

        .. code::

            import fanc

            re_fragments = fanc.genome_regions("hg19_chr18_19.fa", "HindIII")
            rp = fanc.ReadPairs()
            rp.add_regions(re_fragments.regions)

            # read pairs are added here from BAM files
            rp_generator = fanc.SamBamReadPairGenerator("output/sam/SRR4271982_chr18_19_1_sort.bam",
                                                        "output/sam/SRR4271982_chr18_19_2_sort.bam")
            rp.add_read_pairs(rp_generator, threads=4)

        :param read_pairs: iterator over tuples of read pairs. Typically
                           instances of :class:`~ReadPairGenerator`
        :param batch_size: Batch size of read pairs sent to fragment info workers
        :param threads: Number of threads for simultaneous fragment info finding
        """
        self._edges_dirty = True
        self._disable_edge_indexes()

        if self._pair_count is None:
            self._pair_count = sum(edge_table._original_len()
                                   for _, edge_table in self._iter_edge_tables())

        start_time = timer()
        chunk_start_time = timer()
        pairs_counter = 0
        edge = {}
        for fi1, fi2 in self._read_pairs_fragment_info(read_pairs, batch_size=batch_size, threads=threads):
            source, sink = int(fi1[2]), int(fi2[2])
            if source > sink:
                source, sink = sink, source
                fi1, fi2 = fi2, fi1

            edge['ix'] = self._pair_count
            edge['source'] = source
            edge['sink'] = sink
            edge['left_read_position'], edge['right_read_position'] = int(fi1[0]), int(fi2[0])
            edge['left_read_strand'], edge['right_read_strand'] = fi1[1], fi2[1]
            edge['left_fragment_start'], edge['right_fragment_start'] = int(fi1[4]), int(fi2[4])
            edge['left_fragment_end'], edge['right_fragment_end'] = int(fi1[5]), int(fi2[5])
            edge['left_fragment_chromosome'], edge['right_fragment_chromosome'] = int(fi1[3]), int(fi2[3])

            self._edge_buffer.add_dict(edge)
            self._pair_count += 1

        logger.info('Done saving read pairs.')

        if isinstance(read_pairs, ReadPairGenerator):
            stats = read_pairs.stats()
            if 'read_filter_stats' not in self.meta:
                self.meta.read_filter_stats = stats
            else:
                self.meta.read_filter_stats = add_dict(self.meta.read_filter_stats, stats)

        self.flush()
        logger.info("Done adding pairs.")

    def _add_pair(self, pair):
        self.add_edge(pair, check_nodes_exist=False, replace=True)
        self._pair_count += 1

    def _pair_from_row(self, row, lazy_pair=None):
        """
        Convert a PyTables row to a FragmentReadPair
        """
        if lazy_pair is not None:
            lazy_pair.left._row = row
            lazy_pair.right._row = row
            lazy_pair.ix = row['ix']
            return lazy_pair
        else:
            fragment1 = GenomicRegion(start=row['left_fragment_start'],
                                      end=row['left_fragment_end'],
                                      chromosome=self._ix_to_chromosome[row['left_fragment_chromosome']],
                                      ix=row['source'])
            fragment2 = GenomicRegion(start=row['right_fragment_start'],
                                      end=row['right_fragment_end'],
                                      chromosome=self._ix_to_chromosome[row['right_fragment_chromosome']],
                                      ix=row['sink'])

            left_read = FragmentRead(fragment1, position=row['left_read_position'],
                                     strand=row['left_read_strand'])
            right_read = FragmentRead(fragment2, position=row['right_read_position'],
                                      strand=row['right_read_strand'])

            return FragmentReadPair(left_read=left_read, right_read=right_read, ix=row['ix'])

    def get_ligation_structure_biases(self, sampling=None, skip_self_ligations=True,
                                      **kwargs):

        """
        Compute the ligation biases (inward and outward to same-strand) of this data set.

        :param sampling: Approximate number of data points to average per point
                         in the plot. If None (default), this will
                         be determined on a best-guess basis.
        :param skip_self_ligations: If True (default), will not consider
                                    self-ligated fragments for assessing
                                    the error rates.
        :param unfiltered: If True, uses all read pairs, even those that do not
                           pass filters, for the ligation error calculation
        :return: tuple with (list of gap sizes between reads, list of matching le type ratios)
        """
        n_pairs = len(self)
        type_same = 0
        type_inward = 1
        type_outward = 2

        def _init_gaps_and_types():
            same_count = 0
            inward_count = 0
            outward_count = 0
            same_fragment_count = 0
            inter_chrm_count = 0
            gaps = []
            types = []

            max_value = sum(et._original_len() for _, et in self._iter_edge_tables())

            with RareUpdateProgressBar(max_value=max_value, silent=config.hide_progressbars,
                                       prefix="Ligation error") as pb:
                for i, pair in enumerate(self.pairs(lazy=True, excluded_filters='all',
                                                    **kwargs)):
                    if pair.is_same_fragment():
                        same_fragment_count += 1
                        if skip_self_ligations:
                            continue
                    if pair.is_same_chromosome():
                        gap_size = pair.get_gap_size()
                        if gap_size > 0:
                            gaps.append(gap_size)
                            if pair.is_outward_pair():
                                types.append(type_outward)
                                outward_count += 1
                            elif pair.is_inward_pair():
                                types.append(type_inward)
                                inward_count += 1
                            else:
                                types.append(type_same)
                                same_count += 1
                    else:
                        inter_chrm_count += 1
                    pb.update(i)

            logger.info("Pairs: %d" % n_pairs)
            logger.info("Inter-chromosomal: {}".format(inter_chrm_count))
            logger.info("Same fragment: {}".format(same_fragment_count))
            logger.info("Same: {}".format(same_count))
            logger.info("Inward: {}".format(inward_count))
            logger.info("Outward: {}".format(outward_count))
            return gaps, types

        def _sort_data(gaps, types):
            points = zip(gaps, types)
            sorted_points = sorted(points)
            return zip(*sorted_points)

        def _calculate_ratios(gaps, types, sampling):
            x = []
            inward_ratios = []
            outward_ratios = []
            bin_sizes = []
            counter = 0
            same_counter = 0
            mids = 0
            outwards = 0
            inwards = 0
            for typ, gap in zip(types, gaps):
                mids += gap
                if typ == type_same:
                    same_counter += 1
                elif typ == type_inward:
                    inwards += 1
                else:
                    outwards += 1
                counter += 1
                if same_counter > sampling:
                    x.append(int(mids / counter))
                    inward_ratios.append(inwards / same_counter)
                    outward_ratios.append(outwards / same_counter)
                    bin_sizes.append(counter)
                    same_counter = 0
                    counter = 0
                    mids = 0
                    outwards = 0
                    inwards = 0
            return list(map(np.array, [x, inward_ratios, outward_ratios, bin_sizes]))

        gaps, types = _init_gaps_and_types()
        # sort data
        gaps, types = _sort_data(gaps, types)
        # best guess for number of data points
        sampling = max(100, int(n_pairs * 0.0025)) if sampling is None else sampling
        logger.debug("Number of data points averaged per point in plot: {}".format(sampling))
        # calculate ratios
        return _calculate_ratios(gaps, types, sampling)

    @staticmethod
    def _auto_dist(dists, ratios, sample_sizes, p=0.05, expected_ratio=0.5):
        """
        Function that attempts to infer sane distances for filtering inward
        and outward read pairs.

        Use with caution, it is almost always preferable to plot the ligation
        error and choose the cutoff manually.

        :param dists: List of distances in bp.
        :param ratios: List of ratios
        """

        def x_prop(p_obs, p_exp, n):
            obs = p_obs * n
            exp = p_exp * n
            p = (obs + exp) / (n * 2)
            return abs((p_exp - p_obs) / np.sqrt(p * (1 - p) * (2 / n)))

        ratios = np.clip(ratios, 0.0, 1.0)
        z_scores = np.array([x_prop(r, expected_ratio, b) for r, b in zip(ratios, sample_sizes)])
        which_valid = z_scores < 1.96
        which_valid_indices = np.argwhere(which_valid).flatten()
        if len(which_valid_indices) > 0:
            return int(dists[which_valid_indices[0]])
        return None

    def filter(self, pair_filter, queue=False, log_progress=not config.hide_progressbars):
        """
        Apply a :class:`~FragmentReadPairFilter` to the read pairs in this object.

        :param pair_filter: :class:`~FragmentReadPairFilter`
        :param queue: If True, does not do the filtering immediately, but
                      queues this filter. All queued filters can then be run
                      at the same time using :func:`~ReadPairs.run_queued_filters`
        :param log_progress:
        :return:
        """
        pair_filter.set_pairs_object(self)

        total = 0
        filtered = 0
        if not queue:
            with RareUpdateProgressBar(max_value=sum(1 for _ in self._edges),
                                       silent=not log_progress,
                                       prefix="Filter") as pb:
                for i, (_, edge_table) in enumerate(self._iter_edge_tables()):
                    stats = edge_table.filter(pair_filter, _logging=False)
                    for key, value in stats.items():
                        if key != 0:
                            filtered += stats[key]
                        total += stats[key]
                    pb.update(i)
            if log_progress:
                logger.debug("Total: {}. Valid: {}".format(total, total - filtered))
        else:
            self._queued_filters.append(pair_filter)

    def run_queued_filters(self, log_progress=not config.hide_progressbars):
        """
        Run queued filters. See :func:`~ReadPairs.filter`

        :param log_progress: If true, process iterating through all edges
                             will be continuously reported.
        """
        total = 0
        filtered = 0
        with RareUpdateProgressBar(max_value=sum(1 for _ in self._edges),
                                   silent=not log_progress,
                                   prefix="Filter") as pb:
            for i, (_, edge_table) in enumerate(self._iter_edge_tables()):
                for f in self._queued_filters:
                    edge_table.queue_filter(f)

                stats = edge_table.run_queued_filters(_logging=False)
                for key, value in stats.items():
                    if key != 0:
                        filtered += stats[key]
                    total += stats[key]
                pb.update(i)
        if log_progress:
            logger.info("Total: {}. Valid: {}".format(total, total - filtered))

        self._queued_filters = []
        self._update_mappability()

    def filter_pcr_duplicates(self, threshold=3, queue=False):
        """
        Convenience function that applies an :class:`~PCRDuplicateFilter`.

        :param threshold: If distance between two alignments is smaller or
                          equal the threshold, the alignments
                          are considered to be starting at the same position
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('PCR duplicates', 'Mask read pairs that are '
                                                           'considered PCR duplicates')
        pcr_duplicate_filter = PCRDuplicateFilter(pairs=self, threshold=threshold, mask=mask)
        self.filter(pcr_duplicate_filter, queue)

    def filter_inward(self, minimum_distance=None, queue=False, **kwargs):
        """
        Convenience function that applies an :class:`~InwardPairsFilter`.

        :param minimum_distance: Minimum distance inward-facing read
                                 pairs must have to pass this filter
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        :param kwargs: Additional arguments to pass
                       to :func:`~ReadPairs.get_ligation_structure_biases`
        """
        if minimum_distance is None:
            dists, inward_ratios, _, bins_sizes = self.get_ligation_structure_biases(**kwargs)
            minimum_distance = self._auto_dist(dists, inward_ratios, bins_sizes)
        if minimum_distance:
            mask = self.add_mask_description('inward ligation error',
                                             'Mask read pairs that are inward facing and < {}bp apart'
                                             .format(minimum_distance))
            logger.info("Filtering out inward facing read pairs < {} bp apart".format(minimum_distance))
            inward_filter = InwardPairsFilter(minimum_distance=minimum_distance, mask=mask)
            self.filter(inward_filter, queue)
        else:
            raise Exception('Could not automatically detect a sane distance threshold for '
                            'filtering inward reads')

    def filter_outward(self, minimum_distance=None, queue=False, **kwargs):
        """
        Convenience function that applies an :class:`~OutwardPairsFilter`.

        :param minimum_distance: Minimum distance outward-facing read
                                 pairs must have to pass this filter
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        :param kwargs: Additional arguments to pass
                       to :func:`~ReadPairs.get_ligation_structure_biases`
        """
        if minimum_distance is None:
            dists, _, outward_ratios, bins_sizes = self.get_ligation_structure_biases(**kwargs)
            minimum_distance = self._auto_dist(dists, outward_ratios, bins_sizes)
        if minimum_distance:
            mask = self.add_mask_description('outward ligation error',
                                             'Mask read pairs that are outward facing and < {}bp apart'
                                             .format(minimum_distance))
            logger.info("Filtering out outward facing read pairs < {} bp apart".format(minimum_distance))
            outward_filter = OutwardPairsFilter(minimum_distance=minimum_distance, mask=mask)
            self.filter(outward_filter, queue)
        else:
            raise Exception('Could not automatically detect a sane distance threshold for filtering outward reads')

    def filter_ligation_products(self, inward_threshold=None, outward_threshold=None, queue=False, **kwargs):
        """
        Convenience function that applies an :class:`~OutwardPairsFilter` and an :class:`~InwardPairsFilter`.

        :param inward_threshold: Minimum distance inward-facing read
                                 pairs must have to pass this filter.
                                 If None, will be inferred from the data
        :param outward_threshold: Minimum distance outward-facing read
                                 pairs must have to pass this filter.
                                 If None, will be inferred from the data
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        :param kwargs: Additional arguments to pass
                       to :func:`~ReadPairs.get_ligation_structure_biases`
        """
        self.filter_inward(inward_threshold, queue=queue, **kwargs)
        self.filter_outward(outward_threshold, queue=queue, **kwargs)

    def filter_re_dist(self, maximum_distance, queue=False):
        """
        Convenience function that applies an :class:`~ReDistanceFilter`.

        :param maximum_distance: Maximum distance a read can have to the
                                 nearest region border (=restriction site)
        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('restriction site distance',
                                         'Mask read pairs where the cumulative distance of reads to '
                                         'the nearest RE site exceeds {}'.format(maximum_distance))
        re_filter = ReDistanceFilter(maximum_distance=maximum_distance, mask=mask)
        self.filter(re_filter, queue)

    def filter_self_ligated(self, queue=False):
        """
        Convenience function that applies an :class:`~SelfLigationFilter`.

        :param queue: If True, filter will be queued and can be executed
                      along with other queued filters using
                      run_queued_filters
        """
        mask = self.add_mask_description('self-ligations',
                                         'Mask read pairs that represent a self-ligated fragment')
        self_ligation_filter = SelfLigationFilter(mask=mask)
        self.filter(self_ligation_filter, queue)

    def __iter__(self):
        """
        Iterate over unfiltered fragment-mapped read pairs.
        """
        return self.pairs(lazy=False)

    def pairs(self, key=None, lazy=False, *args, **kwargs):
        """
        Iterate over the :class:`~FragmentReadPair` objects.

        :param key: Region string of the form <chromosome>[:<start>-<end>],
                    :class:`~fanc.GenomicRegion` or tuples thereof
        :param lazy: If True, use lazy loading of objects and their attributes.
                     Much faster, but can lead to unexpected results if one is
                     not careful. For example, this: :code:`list(object.pairs())`
                     is not the same as :code:`list(object.pairs(lazy=True))`! In
                     the latter case, all objects in the list will be identical
                     due to the lazy loading process. Only use lazy loading to access
                     attributes in an iterator!
        :param args: Positional arguments passed to :func:`~RegionPairs.edges_dict`
        :param kwargs: Keyword arguments passed to :func:`~RegionPairs.edges_dict`
        :return:
        """
        if lazy:
            fr1 = LazyFragmentRead({}, self, side='left')
            fr2 = LazyFragmentRead({}, self, side='right')
            lazy_pair = FragmentReadPair(fr1, fr2, ix=None)
        else:
            lazy_pair = None
        for row in self.edges_dict(key=key, *args, **kwargs):
            yield self._pair_from_row(row, lazy_pair=lazy_pair)

    def get_edge(self, item, *row_conversion_args, **row_conversion_kwargs):
        """
        Get an edge by index.

        :param row_conversion_args: Arguments passed to :func:`RegionPairs._row_to_edge`
        :param row_conversion_kwargs: Keyword arguments passed to :func:`RegionPairs._row_to_edge`
        :return: :class:`~Edge`
        """
        if item < 0:
            item += len(self)

        l = 0
        for _, edge_table in self._iter_edge_tables():
            if l <= item < l + len(edge_table):
                res = edge_table[item - l]
                return self._row_to_edge(res, *row_conversion_args, **row_conversion_kwargs)
            l += len(edge_table)
        raise IndexError("index out of range (%d)" % item)

    def __getitem__(self, item):
        if isinstance(item, int):
            edge = self.get_edge(item)
            fragment1 = GenomicRegion(start=edge.left_fragment_start,
                                      end=edge.left_fragment_end,
                                      chromosome=self._ix_to_chromosome[edge.left_fragment_chromosome],
                                      ix=edge.source)
            fragment2 = GenomicRegion(start=edge.right_fragment_start,
                                      end=edge.right_fragment_end,
                                      chromosome=self._ix_to_chromosome[edge.right_fragment_chromosome],
                                      ix=edge.sink)

            left_read = FragmentRead(fragment1, position=edge.left_read_position,
                                     strand=edge.left_read_strand)
            right_read = FragmentRead(fragment2, position=edge.right_read_position,
                                      strand=edge.right_read_strand)

            return FragmentReadPair(left_read=left_read, right_read=right_read, ix=edge.ix)
        else:
            pairs = []
            for row in self.edges.get_row_range(item):
                pairs.append(self._pair_from_row(row))
            return pairs

    def __len__(self):
        l = 0
        for _, edge_table in self._iter_edge_tables():
            l += len(edge_table)
        return l

    def to_hic(self, file_name=None, tmpdir=None, _hic_class=Hic):
        """
        Convert this :class:`~ReadPairs` to a :class:`~fanc.Hic` object.

        :param file_name: Path to the :class:`~fanc.Hic` output file
        :param tmpdir: If True (or path to temporary directory) will
                       work in temporary directory until closed
        """
        hic = _hic_class(file_name=file_name, mode='w', tmpdir=tmpdir)
        hic.add_regions(self.regions(), preserve_attributes=False)

        hic._disable_edge_indexes()

        n_pairs = len(self)
        pairs_counter = 0
        with RareUpdateProgressBar(max_value=n_pairs, silent=config.hide_progressbars,
                                   prefix="Hi-C convert") as pb:
            for _, pairs_edge_table in self._iter_edge_tables():

                partition_edge_buffer = defaultdict(lambda: defaultdict(int))
                for row in pairs_edge_table:
                    key = (row['source'], row['sink'])
                    source_partition = self._get_partition_ix(key[0])
                    sink_partition = self._get_partition_ix(key[1])
                    partition_edge_buffer[(source_partition, sink_partition)][key] += 1
                    pb.update(pairs_counter)
                    pairs_counter += 1

                for hic_partition_key, edge_buffer in viewitems(partition_edge_buffer):
                    hic_edge_table = hic._edge_table(hic_partition_key[0], hic_partition_key[1])
                    row = hic_edge_table.row

                    for (source, sink), weight in viewitems(edge_buffer):
                        row['source'] = source
                        row['sink'] = sink
                        row[hic._default_score_field] = float(weight)
                        row.append()
                    hic_edge_table.flush(update_index=False)
        hic.flush()

        hic._enable_edge_indexes()

        return hic

    def pairs_by_chromosomes(self, chromosome1, chromosome2, **kwargs):
        """
        Only iterate over read pairs in this combination of chromosomes.
        :param chromosome1: Name of first chromosome
        :param chromosome2: Name of second chromosome
        :param kwargs: Keyword arguments passed to :func:`~ReadPairs.pairs`
        :return:
        """
        return self.pairs(key=(chromosome1, chromosome2), **kwargs)

    def filter_statistics(self):
        """
        Get filtering statistics for this object.
        :return: dict with {filter_type: count, ...}
        """
        try:
            read_stats = self.meta.read_filter_stats
        except AttributeError:
            read_stats = dict()

        pair_stats = self.mask_statistics(self._pairs)
        if 'valid' in pair_stats:
            read_stats['valid'] = pair_stats['valid']
        pair_stats.update(read_stats)
        return pair_stats


class FragmentReadPair(object):
    """
    Container for two paired :class:`~FragmentRead` objects.
    """

    def __init__(self, left_read, right_read, ix=None):
        self.left = left_read
        self.right = right_read
        self.ix = ix

    def is_same_chromosome(self):
        """
        Check if both reads are mapped to the same chromosome.

        :return: True is reads map to the same chromosome, False
                 otherwise
        """
        return self.left.fragment.chromosome == self.right.fragment.chromosome

    def is_inward_pair(self):
        """
        Check if reads form an "inward-facing" pair.

        A pair is considered inward-facing if the left read maps
        to the plus the right read to the minus strand and both
        reads map to the same chromosome.

        :return: True if reads are inward-facing, False otherwise
        """
        if not self.is_same_chromosome():
            return False

        if self.left.strand == 1 and self.right.strand == -1:
            return True
        return False

    def is_outward_pair(self):
        """
        Check if reads form an "outward-facing" pair.

        A pair is considered outward-facing if the left read maps
        to the minus the right read to the plus strand and both
        reads map to the same chromosome.

        :return: True if reads are outward-facing, False otherwise
        """
        if not self.is_same_chromosome():
            return False

        if self.left.strand == -1 and self.right.strand == 1:
            return True
        return False

    def is_same_pair(self):
        """
        Check if reads face in the same direction.

        :return: True if reads are facing in the same direction,
                 False otherwise.
        """
        if not self.is_same_chromosome():
            return False

        if self.left.strand == self.right.strand:
            return True
        return False

    def is_same_fragment(self):
        """
        Check if reads map to the same fragment.

        :return: True if reads map to the same fragment,
                 False otherwise.
        """
        if not self.is_same_chromosome():
            return False

        return self.left.fragment.start == self.right.fragment.start

    def get_gap_size(self):
        """
        Get the gap size in base pairs between the fragments these
        reads map to.

        :return: 0 if reads map to the same fragment or neighboring
                 fragments, the distance between fragments if they
                 are on the same chromosome, None otherwise
        """
        if not self.is_same_chromosome():
            return None

        if self.is_same_fragment():
            return 0

        gap = self.right.fragment.start - self.left.fragment.end

        if gap == 1:  # neighboring fragments
            return 0

        return gap

    def __getitem__(self, key):
        if key == 0:
            return self.left
        if key == 1:
            return self.right
        raise KeyError("Can only access read [0] and read [1], not '{}'".format(key))

    def __repr__(self):
        left_repr = self.left.__repr__()
        right_repr = self.right.__repr__()
        return "{} -- {}".format(left_repr, right_repr)


class ReadFilter(object):
    """
    Abstract class that provides filtering functionality for
    :class:`~MinimalRead`, :class:`~pysam.AlignedSegment` or
    compatible.

    To create a custom :class:`~ReadFilter`, extend this
    class and override the valid_read(self, read) method.
    valid_read should return False for a specific read object
    if the object is supposed to be filtered/masked and True
    otherwise. See :class:`~QualityFilter` for an example.

    Pass a custom filter to the filter method in
    :class:`~ReadPairGenerator` to apply it.
    """

    def __init__(self, mask=None):
        """
        Initialize ReadFilter.

        :param mask: The :class:`~fanc.general.Mask` object that
                     should be used to mask
                     filtered Read objects. If None the default
                     Mask will be used.
        """
        if mask is not None:
            if isinstance(mask, Mask):
                self.mask = mask
            elif isinstance(mask, string_types):
                self.mask = Mask(mask, mask)
        else:
            self.mask = Mask('default', 'Default mask')

    def valid_read(self, read):
        """
        Determine if a read is valid or should be filtered.

        When implementing custom read filters this method must be
        overridden. It should return False for reads that
        are to be filtered and True otherwise.

        Internally, the ReadPairs object will iterate over all Read
        instances to determine their validity on an individual
        basis.

        :param read: A :class:`~Read` object
        :return: True if Read is valid, False otherwise
        """
        raise NotImplementedError("ReadFilters must implement valid_read function")


class QualityFilter(ReadFilter):
    """
    Filter mapped reads based on mapping quality.
    """

    def __init__(self, cutoff=30, mask=None):
        """
        :param cutoff: Lowest mapping quality that is still
                       considered acceptable for a mapped read.
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        super(QualityFilter, self).__init__(mask)
        self.cutoff = cutoff

    def valid_read(self, read):
        """
        Check if a read has a mapq >= cutoff.
        """
        return read.mapq >= self.cutoff

    def __reduce__(self):
        return QualityFilter, (self.cutoff, self.mask)


class ContaminantFilter(ReadFilter):
    """
    Filter reads that also map to a contaminant genome.
    """

    def __init__(self, contaminant_reads, mask=None):
        """
        :param contaminant_reads: A SAM/BAM file representing a
                                  contaminant
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        super(ContaminantFilter, self).__init__(mask)

        self.contaminant_reads = contaminant_reads
        self.contaminant_names = set()
        with pysam.AlignmentFile(contaminant_reads) as contaminant:
            for read in contaminant:
                self.contaminant_names.add(read.qname)

    def valid_read(self, read):
        """
        Check if a read also maps to a contaminant
        """
        if read.qname in self.contaminant_names:
            return False
        return True

    def __reduce__(self):
        return ContaminantFilter, (self.contaminant_reads, self.mask)


class BwaMemQualityFilter(ReadFilter):
    """
    Filters :code:`bwa mem` generated alignments
    based on the alignment score (normalized by
    the length of the alignment).
    """
    def __init__(self, cutoff=0.90, mask=None):
        """
        :param cutoff: Ratio of the alignment score to the maximum score
                       possible for an alignment that long
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        super(BwaMemQualityFilter, self).__init__(mask)
        self.cutoff = cutoff

    def valid_read(self, read):
        """
        Check if a read has a high alignment score.
        """
        if read.alen:
            return float(read.get_tag('AS')) / read.alen >= self.cutoff
        return False

    def __reduce__(self):
        return BwaMemQualityFilter, (self.cutoff, self.mask)


class UniquenessFilter(ReadFilter):
    """
    Filter reads that do not map uniquely to the reference sequence.
    """

    def __init__(self, strict=True, mask=None):
        """
        :param strict: If True, valid_read checks only for the
                       presence of an XS tag. If False, the value
                       of an XS tag also has to be different from 0.
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        self.strict = strict
        super(UniquenessFilter, self).__init__(mask)

    def valid_read(self, read):
        """
        Check if a read has an XS tag.

        If strict is enabled checks if a read has an XS tag.
        If not strict, XS has to be smaller than AS (alignment score)
        for a valid read.
        """
        try:
            tag_xs = read.get_tag('XS')
            if self.strict:
                return False
            else:
                tag_as = read.get_tag('AS')
                if tag_as <= tag_xs:
                    return False
        except KeyError:
            pass
        return True

    def __reduce__(self):
        return UniquenessFilter, (self.strict, self.mask)


class BwaMemUniquenessFilter(ReadFilter):
    """
    Filters `bwa mem` generated alignments based on whether they are unique or not.

    """
    def __init__(self, strict=False, mask=None):
        """
        :param strict: If True, valid_read checks only for the
                       presence of an XA tag. If False, the edit
                       distance (NM) of an alternative alignment has to be
                       the same or better as the original NM.
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        super(BwaMemUniquenessFilter, self).__init__(mask)
        self.strict = strict

    def valid_read(self, read):
        try:
            xa = read.get_tag('XA')
            if self.strict:
                return False

            try:
                nm = read.get_tag('NM')
            except KeyError:
                return False

            for alt in xa.split(';'):
                if alt == '':
                    continue
                _, _, _, nm_alt = alt.split(',')
                if int(nm_alt) <= nm:
                    return False
        except KeyError:
            pass
        return True

    def __reduce__(self):
        return BwaMemUniquenessFilter, (self.strict, self.mask)


class UnmappedFilter(ReadFilter):
    """
    Filter reads that do not map to the reference sequence.
    """
    def __init__(self, mask=None):
        """
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        super(UnmappedFilter, self).__init__(mask)

    def valid_read(self, read):
        """
        Check if the the flag bit 4 is set.
        """
        if read.flag & 4:
            return False
        return True

    def __reduce__(self):
        return UnmappedFilter, (self.mask,)


class FragmentReadPairFilter(with_metaclass(ABCMeta, MaskFilter)):
    """
    Abstract class that provides filtering functionality for the
    :class:`~FragmentReadPair` object.

    Extends MaskFilter and overrides valid(self, read) to make
    :class:`~FragmentReadPair` filtering more "natural".

    To create custom filters for the :class:`~FragmentMappedReadPairs`
    object, extend this
    class and override the :func:`~FragmentMappedReadPairFilter.valid_pair` method.
    valid_pair should return False for a specific :class:`~FragmentReadPair` object
    if the object is supposed to be filtered/masked and True
    otherwise. See :class:`~InwardPairsFilter` for an example.

    Pass a custom filter to the filter method in :class:`~FragmentMappedReadPairs`
    to apply it.
    """

    def __init__(self, mask=None):
        super(FragmentReadPairFilter, self).__init__(mask)
        self.pairs = None
        self._lazy_pair = None

    def set_pairs_object(self, pairs):
        self.pairs = pairs
        fr1 = LazyFragmentRead({}, pairs, side='left')
        fr2 = LazyFragmentRead({}, pairs, side='right')
        self._lazy_pair = FragmentReadPair(fr1, fr2)

    @abstractmethod
    def valid_pair(self, fr_pair):
        pass

    def valid(self, row):
        """
        Map validity check of rows to pairs.
        """
        pair = self.pairs._pair_from_row(row, lazy_pair=self._lazy_pair)
        return self.valid_pair(pair)


class InwardPairsFilter(FragmentReadPairFilter):
    """
    Filter inward-facing read pairs at a distance less
    than a specified cutoff.
    """

    def __init__(self, minimum_distance=10000, mask=None):
        """
        Initialize filter.

        :param minimum_distance: Minimum distance below which
                                 reads are invalidated
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        super(InwardPairsFilter, self).__init__(mask=mask)
        self.minimum_distance = minimum_distance

    def valid_pair(self, pair):
        """
        Check if a pair is inward-facing and <minimum_distance> apart.
        """
        if pair.is_inward_pair() and pair.get_gap_size() <= self.minimum_distance:
            return False
        return True


class PCRDuplicateFilter(FragmentReadPairFilter):
    """
    Masks alignments that are suspected to be PCR duplicates.
    In order to be considered duplicates, two pairs need to have identical
    start positions of their respective left alignments AND of their right alignments.
    """

    def __init__(self, pairs, threshold=2, mask=None):
        """
        Initialize filter with filter settings.

        :param pairs: The :class:`~FragmentReadPairs` instance that the filter will be
                      applied to
        :param threshold: If distance between two alignments is smaller or equal the threshold,
                          the alignments are considered to be starting at the same position
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        FragmentReadPairFilter.__init__(self, mask=mask)
        self.threshold = threshold
        self.pairs = pairs
        self.duplicates_set = set()
        self.duplicate_stats = defaultdict(int)
        original_len = 0
        for _, edge_table in self.pairs._iter_edge_tables():
            original_len += edge_table._original_len()
            self._mark_duplicates(edge_table)

        n_dups = len(self.duplicates_set)
        percent_dups = 1. * n_dups / original_len
        logger.info("PCR duplicate stats: " +
                    "{} ({:.1%}) of pairs marked as duplicate. ".format(n_dups, percent_dups) +
                    " (multiplicity:occurances) " +
                    " ".join("{}:{}".format(k, v) for k, v in self.duplicate_stats.items()))

    def _mark_duplicates(self, edge_table):
        pairs = [
            (
                row['ix'],
                row['left_fragment_chromosome'], row['right_fragment_chromosome'],
                row['left_read_position'], row['right_read_position']
             ) for row in edge_table._iter_visible_and_masked()
        ]
        pairs = sorted(pairs, key=lambda p: p[3])

        current_positions = {}
        current_duplicates = {}
        for ix, left_chromosome, right_chromosome, left_position, right_position in pairs:
            chromosomes = (left_chromosome, right_chromosome)

            # case 1: no current duplicates
            if current_positions.get(chromosomes) is None:
                current_positions[chromosomes] = (left_position, right_position)
                current_duplicates[chromosomes] = 1
                continue

            # case 2: found duplicate
            if (abs(left_position - current_positions[chromosomes][0]) <= self.threshold and
                    abs(right_position - current_positions[chromosomes][1]) <= self.threshold):
                self.duplicates_set.add(ix)
                current_duplicates[chromosomes] += 1
                continue

            # update statistics
            if current_duplicates[chromosomes] > 1:
                self.duplicate_stats[current_duplicates[chromosomes]] += 1

            current_positions[chromosomes] = (left_position, right_position)
            current_duplicates[chromosomes] = 1

    def valid_pair(self, pair):
        """
        Check if a pair is duplicated.
        """
        if pair.ix in self.duplicates_set:
            return False
        return True


class OutwardPairsFilter(FragmentReadPairFilter):
    """
    Filter outward-facing read pairs at a distance less
    than a specified cutoff.
    """

    def __init__(self, minimum_distance=10000, mask=None):
        """
        Initialize filter with filter settings.

        :param minimum_distance: Minimum distance below which
                                 outward-facing reads are invalidated
        :param mask: Optional Mask object describing the mask
                     that is applied to filtered reads.
        """
        super(OutwardPairsFilter, self).__init__(mask=mask)
        self.minimum_distance = minimum_distance

    def valid_pair(self, pair):
        if not pair.is_outward_pair():
            return True

        if pair.get_gap_size() > self.minimum_distance:
            return True
        return False


class ReDistanceFilter(FragmentReadPairFilter):
    """
    Filters read pairs where the sum of distances of each read to its nearest
    restriction site is larger than a certain distance.
    """

    def __init__(self, maximum_distance=10000, mask=None):
        super(ReDistanceFilter, self).__init__(mask=mask)
        self.maximum_distance = maximum_distance

    def valid_pair(self, pair):
        """
        Check sum of restriction site distances < maximum_distance.
        """
        d1 = min(abs(pair.left.position - pair.left.fragment.start),
                 abs(pair.left.position - pair.left.fragment.end))
        d2 = min(abs(pair.right.position - pair.right.fragment.start),
                 abs(pair.right.position - pair.right.fragment.end))

        if d1 + d2 > self.maximum_distance:
            return False

        return True


class SelfLigationFilter(FragmentReadPairFilter):
    """
    Filters read pairs where one or both reads map to the
    same restriction fragment.
    """

    def __init__(self, mask=None):
        super(SelfLigationFilter, self).__init__(mask=mask)

    def valid_pair(self, pair):
        """
        Check if both reads are on the same fragment.
        """
        if pair.is_same_fragment():
            return False
        return True



import threading
import logging
import warnings
import os

from genomic_regions import load as gr_load
from ..registry import class_id_dict
import tables

logger = logging.getLogger(__name__)

fanc_access_lock = threading.Lock()


def load(file_name, *args, **kwargs):
    """
    Load a file into your current Python session.

    :func:`~load` is a magic function that replaces the need for importing
    files using different classes or functions. It "just works" for all
    objects generated by FAN-C (:class:`~Hic`, :class:`~ReadPairs`,
    :class:`~ABCompartmentMatrix`, ...), for compatible Hi-C files from
    `Cooler <https://github.com/mirnylab/cooler>`_ or
    `Juicer <https://github.com/aidenlab/juicer>`_, and most of the major
    file formats for genomic regions (BED, GFF, BigWig, Tabix, ...).

    Simply run

    .. code::

        o = fanc.load("/path/to/file")

    Depending on the file type, the returned object can be the instance of
    one (or more) of these classes:

    - :class:`~RegionBased` for genomic region formats (BED, GFF, ..., but
      also most FAN-C objects)
    - :class:`~RegionMatrixContainer` or :class:`~RegionPairsContainer` for
      read pair or matrix-based FAN-C objects, as well as Cooler and Juicer
      files
    - :class:`~pysam.AlignmentFile` for SAM/BAM files

    :param file_name: Path to file
    :param args: Positional arguments passed to the class/function that can
                 load the file
    :param kwargs: Keyword arguments passed to the class/function that can
                   load the file
    :return: object (:class:`~RegionBased`, :class:`~RegionMatrixContainer`,
             :class:`~RegionPairsContainer`, or :class:`~pysam.AlignmentFile`)
    """
    mode = kwargs.pop('mode', 'r')
    file_name = os.path.expanduser(file_name)
    logger.debug("Searching for file type compatible with {}".format(file_name))

    try:
        logger.debug("Trying FileBased classes")

        f = tables.open_file(file_name, mode='r')
        try:
            classid = f.get_node('/', 'meta_information').meta_node.attrs['_classid']
            classid = classid.decode() if isinstance(classid, bytes) else classid
        finally:
            f.close()
        logger.debug("Class ID string: {}".format(classid))
        cls_ = class_id_dict[classid]
        logger.debug("Detected {}".format(cls_))
        return cls_(file_name=file_name, mode=mode, *args, **kwargs)
    except (tables.HDF5ExtError, AttributeError, KeyError) as e:
        logger.debug("Not a FileBased class (exception: {})".format(e))
    except OSError as e:
        logger.debug("Exact filename not found, might still be cooler uri (exception: {})".format(e))

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from fanc.compatibility.cooler import is_cooler, CoolerHic
        if is_cooler(file_name):
            logger.debug("Cooler file detected")
            return CoolerHic(file_name, *args, **kwargs)
    except (ImportError, OSError, FileNotFoundError) as e:
        logger.debug("Not a Cooler file (exception: {}".format(e))

    from fanc.compatibility.juicer import JuicerHic, is_juicer
    if is_juicer(file_name):
        return JuicerHic(file_name, *args, **kwargs)
    else:
        logger.debug("Not a Juicer file")

    logger.debug("Passing the file on to genomic_region.load to try region_based file types")
    return gr_load(file_name, *args, **kwargs)



import numpy as np
from bisect import bisect_left
from collections import defaultdict
import intervaltree
from genomic_regions import GenomicRegion
from future.utils import string_types


def vector_enrichment_profile(matrix, vector, mappable=None, per_chromosome=True,
                              percentiles=(20.0, 40.0, 60.0, 80.0, 100.0),
                              symmetric_at=None, exclude_chromosomes=(),
                              intra_chromosomal=True, inter_chromosomal=False,
                              collapse_identical_breakpoints=False):
    if len(exclude_chromosomes) > 0:
        chromosome_bins = matrix.chromosome_bins
        exclude_vector = []
        for chromosome in matrix.chromosomes():
            if chromosome not in exclude_chromosomes:
                b = chromosome_bins[chromosome]
                for v in vector[b[0]:b[1]]:
                    exclude_vector.append(v)
                # exclude_vector += vector[b[0]:b[1]]
    else:
        exclude_vector = vector
    exclude_vector = np.array(exclude_vector)

    if symmetric_at is not None:
        lv = exclude_vector[exclude_vector <= symmetric_at]
        gv = exclude_vector[exclude_vector > symmetric_at]
        lv_cutoffs = np.nanpercentile(lv, percentiles)
        gv_cutoffs = np.nanpercentile(gv, percentiles)
        bin_cutoffs = np.concatenate((lv_cutoffs, gv_cutoffs))
    else:
        bin_cutoffs = np.nanpercentile(exclude_vector, percentiles)

    if collapse_identical_breakpoints:
        new_bin_cutoffs = []
        for i in range(1, len(bin_cutoffs)):
            if not np.isclose(bin_cutoffs[i], bin_cutoffs[i-1]):
                new_bin_cutoffs.append(bin_cutoffs[i])
        bin_cutoffs = new_bin_cutoffs

    if isinstance(intra_chromosomal, bool):
        if intra_chromosomal:
            intra_chromosomal = 0
        else:
            intra_chromosomal = -1

    bins = []
    for value in vector:
        bins.append(bisect_left(bin_cutoffs, value))

    s = len(bin_cutoffs)
    m = np.zeros((s, s))
    c = np.zeros((s, s))

    bin_size = matrix.bin_size
    if mappable is None:
        mappable = matrix.mappable()

    if per_chromosome:
        chromosomes = matrix.chromosomes()
        for chr1_ix in range(len(chromosomes)):
            chromosome1 = chromosomes[chr1_ix]
            if chromosome1 in exclude_chromosomes:
                continue

            for chr2_ix in range(chr1_ix, len(chromosomes)):
                chromosome2 = chromosomes[chr2_ix]
                if chromosome2 in exclude_chromosomes:
                    continue

                is_intra_chromosomal = chromosome1 == chromosome2

                if not inter_chromosomal and not is_intra_chromosomal:
                    continue

                oem = matrix.matrix((chromosome1, chromosome2),
                                    oe=True, oe_per_chromosome=True)
                for i, row_region in enumerate(oem.row_regions):
                    if not mappable[row_region.ix]:
                        continue
                    i_bin = s - bins[row_region.ix] - 1
                    for j, col_region in enumerate(oem.col_regions):
                        if not mappable[col_region.ix]:
                            continue

                        if is_intra_chromosomal and abs(j - i) * bin_size < intra_chromosomal:
                            continue

                        j_bin = s - bins[col_region.ix] - 1
                        value = oem[i, j]

                        m[i_bin, j_bin] += value
                        c[i_bin, j_bin] += 1
                        m[j_bin, i_bin] += value
                        c[j_bin, i_bin] += 1
    else:
        oem = matrix.matrix(oe=True)
        for i in range(oem.shape[0]):
            if not mappable[i]:
                continue
            i_bin = s - bins[i] - 1
            for j in range(i, oem.shape[1]):
                if not mappable[j]:
                    continue

                is_intra_chromosomal = oem.row_regions[i].chromosome == oem.col_regions[j].chromosome

                if is_intra_chromosomal:
                    if abs(j - i) * bin_size < intra_chromosomal:
                        continue
                elif not inter_chromosomal:
                    continue

                j_bin = s - bins[j] - 1
                value = oem[i, j]

                m[i_bin, j_bin] += value
                c[i_bin, j_bin] += 1
                m[j_bin, i_bin] += value
                c[j_bin, i_bin] += 1

    m /= c
    # m[c == 0] = 0
    rev_cutoffs = bin_cutoffs[::-1]
    for i in range(len(rev_cutoffs) - 1, 0, -1):
        if np.isclose(rev_cutoffs[i - 1], rev_cutoffs[i]):
            m[:, i - 1] = m[:, i]
            m[i - 1, :] = m[i, :]

    return np.log2(m), rev_cutoffs


class RegionScoreMatrix(np.ma.MaskedArray):
    def __new__(cls, input_matrix, row_regions=None, columns=None,
                mask=None, *args, **kwargs):
        obj = np.asarray(input_matrix).view(cls, *args, **kwargs)
        obj._row_region_trees = None
        obj.row_regions = None
        obj.columns = columns
        obj.set_row_regions(row_regions)
        obj.mask = mask
        return obj

    def _interval_tree_regions(self, regions):
        intervals = defaultdict(list)
        for i, region in enumerate(regions):
            interval = intervaltree.Interval(region.start - 1, region.end,
                                             data=i)
            intervals[region.chromosome].append(interval)

        interval_trees = {chromosome: intervaltree.IntervalTree(intervals)
                          for chromosome, intervals in intervals.items()}
        return interval_trees

    def set_row_regions(self, regions):
        self.row_regions = regions
        if regions is not None:
            self._row_region_trees = self._interval_tree_regions(regions)
        else:
            self._row_region_trees = None

    def __array_finalize__(self, obj):
        if isinstance(self, np.ma.core.MaskedArray):
            np.ma.MaskedArray.__array_finalize__(self, obj)

        if obj is None:
            return

        self.set_row_regions(getattr(obj, 'row_regions', None))
        self.columns = getattr(obj, 'columns', None)

    def __setitem__(self, key, item):
        self._setitem = True
        try:
            if isinstance(self, np.ma.core.MaskedArray):
                np.ma.MaskedArray.__setitem__(self, key, item)
            else:
                np.ndarray.__setitem__(self, key, item)
        finally:
            self._setitem = False

    def __getitem__(self, index):
        self._getitem = True

        try:
            all_cols = slice(0, len(self.columns), 1)
        except (AttributeError, TypeError):
            try:
                all_cols = slice(0, self.shape[1], 1)
            except IndexError:
                all_cols = None

        # convert string types into region indexes
        if isinstance(index, tuple):
            if len(index) == 2:
                row_key = self._convert_key(
                    index[0],
                    self._row_region_trees if hasattr(self, '_row_region_trees') else None
                )
                col_key = index[1]
                index = (row_key, col_key)
            elif len(index) == 1:
                row_key = self._convert_key(index[0],
                                            self._row_region_trees
                                            if hasattr(self, '_row_region_trees') else None)

                col_key = all_cols
                index = (row_key, )
            else:
                col_key = all_cols
                row_key = index
                index = row_key
        else:
            row_key = self._convert_key(index,
                                        self._row_region_trees
                                        if hasattr(self, '_row_region_trees') else None)

            col_key = all_cols
            index = row_key

        try:
            if isinstance(self, np.ma.core.MaskedArray):
                out = np.ma.MaskedArray.__getitem__(self, index)
            else:
                out = np.ndarray.__getitem__(self, index)
        finally:
            self._getitem = False

        if not isinstance(out, np.ndarray):
            return out

        # get regions
        try:
            row_regions = self.row_regions[row_key]
        except TypeError:
            row_regions = None

        try:
            columns = self.columns[col_key]
        except TypeError:
            columns = None

        if isinstance(row_regions, GenomicRegion):
            out.row_regions = [row_regions]
        else:
            out.row_regions = row_regions

        if not isinstance(columns, list):
            out.columns = [columns]
        else:
            out.columns = columns

        return out

    def __getslice__(self, start, stop):
        return self.__getitem__(slice(start, stop))

    def _convert_key(self, key, region_trees):
        if isinstance(key, string_types):
            key = GenomicRegion.from_string(key)

        if isinstance(key, GenomicRegion):
            start = None
            stop = None
            try:
                key_start = 0 if key.start is None else max(0, key.start - 1)
                key_end = key.end
                for interval in region_trees[key.chromosome][key_start:key_end]:
                    i = interval.data
                    start = min(i, start) if start is not None else i
                    stop = max(i + 1, stop) if stop is not None else i + 1
            except KeyError:
                raise ValueError("Requested chromosome {} was not "
                                 "found in this matrix.".format(key.chromosome))

            if start is None or stop is None:
                raise ValueError("Requested region {} was not found in this matrix.".format(key))

            return slice(start, stop, 1)
        return key


import itertools
try:
    from itertools import izip as zip
except ImportError:
    pass
import random
import collections
import progressbar
import tables as t
import re
import os
import errno
from builtins import object
from datetime import datetime
from Bio import Restriction
from Bio.Seq import reverse_complement
from future.utils import string_types
import pysam
import threading
import multiprocessing as mp
import warnings

random.seed(0)

sam_reference_consumers = {0, 2, 3, 7, 8}  # M, D, N, =, X
sam_query_consumers = {0, 1, 4, 7, 8}  # M, I, S, =, X


class Map(dict):
    """
    Example:
    m = Map({'first_name': 'Foo'}, last_name='Bar', age=99, sports=['Baz'])
    """
    def __init__(self, *args, **kwargs):
        super(Map, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Map, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Map, self).__delitem__(key)
        del self.__dict__[key]


class CachedIterator(object):
    def __init__(self, it, cache_size=1):
        self.it = iter(it)
        self.current = None
        self.cache = collections.deque(maxlen=cache_size)
        self.cache_offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        try:
            if self.cache_offset:
                self.cache_offset -= 1
                if self.cache_offset:
                    return self.cache[-self.cache_offset]
                return self.current
            self.cache.append(self.current)
            self.current = next(self.it)
            return self.current
        except StopIteration:
            return None

    def prev(self):
        try:
            self.cache_offset += 1
            return self.cache[-self.cache_offset]
        except IndexError as e:
            self.cache_offset -= 1
            raise e


def ranges(i):
    for _, b in itertools.groupby(enumerate(i), lambda xy: xy[1] - xy[0]):
        b = list(b)
        yield b[0][1], b[-1][1]


def range_overlap(source_start, source_end, sink_start, sink_end):
    overlap = (max(source_start, sink_start), min(source_end, sink_end))
    if overlap[0] <= overlap[1]:
        return overlap
    return None


def bit_flags_from_int(number, base=2):
    # find largest power of 2
    exponent = -1
    while base**(1+exponent) <= number:
        exponent += 1
    
    # number does not contain any power of 2 (0)
    if exponent == -1:
        return set([])
    
    # find all powers
    bits = []
    while exponent >= 0:
        power = base**exponent
        if number - power >= 0:
            number -= power
            bits.append(exponent)
        exponent -= 1
    
    return set(bits)


def find_alignment_match_positions(alignment, longest=False):
    reference_pos = alignment.reference_start
    query_pos = 0
    reference_matches = []
    query_matches = []
    longest_match_ix = None
    longest_match_length = 0
    try:
        for operation, n in alignment.cigartuples:
            # longest actual match
            if operation == 0 or operation == 7:
                match_start_ref, match_end_ref = reference_pos, reference_pos + n
                match_start_query, match_end_query = query_pos, query_pos + n
                reference_matches.append((match_start_ref, match_end_ref))
                query_matches.append((match_start_query, match_end_query))
                if longest_match_length < n:
                    longest_match_ix = len(reference_matches) - 1
                    longest_match_length = n

            if operation in sam_reference_consumers:
                reference_pos += n

            if operation in sam_query_consumers:
                query_pos += n

        if longest:
            return reference_matches[longest_match_ix], query_matches[longest_match_ix]
        else:
            return reference_matches, query_matches
    except TypeError:
        if alignment.is_unmapped:
            warnings.warn("Cannot extract match positions for {} because it is unmapped!".format(alignment.qname))
        else:
            warnings.warn("Cannot extract match positions for {} (unknown reason)!".format(alignment.qname))
        return (None, None), (None, None)


def distribute_integer(value, divisor, _shuffle=True):
    a = [int(value/divisor)] * divisor
    remaining = value - sum(a)
    i = 0
    while remaining:
        a[i] += 1
        i += 1
        remaining -= 1
    if _shuffle:
        random.shuffle(a)
    return a


def create_col_index(col):
    """
    Create index for Pytables column in a
    safe manner.
    If index already exists, does nothing.
    If index is corrupt, recreates index.
    """
    try:
        col.create_index()
    except ValueError:
        # Index exists
        pass
    except t.NodeError:
        col.remove_index()
        col.create_index()


def to_slice(l):
    if len(l) == 0:
        return slice(0, 0)
    if len(l) == 1:
        return slice(l[0], l[0]+1)
    if len(l) == 2:
        return slice(l[0], l[1]+1, l[1]-l[0])

    d = l[1] - l[0]
    for i in range(len(l)-1):
        if l[i+1] - l[i] != d:
            raise ValueError("Step size between elements varies, cannot make slice")
    return slice(l[0], l[-1]+1, d)


def mkdir(*dir_name):
    dir_name = os.path.expanduser(os.path.join(*dir_name))

    try:
        os.makedirs(dir_name)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

    if not dir_name.endswith('/'):
        dir_name += '/'

    return dir_name


def which(program):
    """
    Check if executable exists in PATH
    :param program: executable name or path to executable
    :return: full path if found, None if not
    """
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def human_format(num, precision=2, lowercase=False):
    """
    Format a number as a string, suffixing letter for 1000 (K), 100000 (M), ...
    :param num: any number larger than zero
    :param precision: number of positions after decimal point
    :param lowercase: return lowercase suffix letters if True
    :return: string representing the number
    """
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0

    num = round(num, precision)

    if num - int(num) == 0:
        num = int(num)
    num_str = '{}{}'.format(num, ['', 'k', 'M', 'G', 'T', 'P'][magnitude])
    return num_str if not lowercase else num_str.lower()


def str_to_int(num_string, decimal_separator='.', thousand_separator=','):
    try:
        num_string = num_string.replace(thousand_separator, '').lower()
    except AttributeError:
        return num_string

    try:
        return int(num_string)
    except ValueError:
        i = 0
        while i < len(num_string) and (num_string[i].isdigit() or num_string[i] == decimal_separator):
            i += 1

        try:
            number = float(num_string[:i])
            suffix = num_string[i:]

            multipliers = {
                'gb': 1000000000,
                'mb': 1000000,
                'kb': 1000,
                'bp': 1,
                'g': 1000000000,
                'm': 1000000,
                'k': 1000,
                'b': 1,
            }

            return int(number * multipliers[suffix])
        except (KeyError, ValueError):
            raise ValueError("Cannot convert '{}' to integer!".format(num_string))


def natural_sort(l):
    def convert(text):
        return int(text) if text.isdigit() else text.lower()
    return sorted(l, key=lambda key: [convert(c) for c in re.split('([0-9]+)', key)])


def natural_cmp(pa, pb):
    i = 0
    j = 0

    if pa[-2] == '/' == pb[-2]:
        pa = pa[:-2]
        pb = pb[:-2]

    try:
        while i < len(pa) and j < len(pb):
            if pa[i].isdigit() and pb[j].isdigit():
                # digit comparison
                while pa[i] == '0':
                    i += 1
                while pb[j] == '0':
                    j += 1
                while pa[i].isdigit() and pb[j].isdigit() and pa[i] == pb[j]:
                    i += 1
                    j += 1
                if pa[i].isdigit() and pb[j].isdigit():
                    k = 0
                    try:
                        while pa[i+k].isdigit() and pb[j+k].isdigit():
                            k += 1
                    except IndexError:
                        if i+k < len(pa):
                            return 1
                        if j+k < len(pb):
                            return -1
                        # both ran over index
                        return int(pa[i]) - int(pb[j])
                    return 1 if pa[i+k].isdigit() else -1 if pb[j+k].isdigit() else int(pa[i]) - int(pb[j])
                elif pa[i].isdigit():
                    return 1
                elif pb[j].isdigit():
                    return -1
                elif i != j:
                    return 1 if i < j else -1
            else:
                # string comparison
                if pa[i] != pb[j]:
                    return -1 if pa[i] < pb[j] else 1
                i += 1
                j += 1
    except IndexError:
        pass
    return 1 if i < len(pa) else -1 if j < len(pb) else 0


def ligation_site_pattern(restriction_enzyme):
    if isinstance(restriction_enzyme, string_types):
        if "^" in restriction_enzyme and "_" in restriction_enzyme:
            cut_pattern = restriction_enzyme
        else:
            restriction_enzyme = getattr(Restriction, restriction_enzyme)
            cut_pattern = restriction_enzyme.elucidate()
    else:
        cut_pattern = restriction_enzyme.elucidate()

    left_side = []
    right_side = []
    for character in cut_pattern:
        if not (len(left_side) > 0 and left_side[-1] == '_'):
            if not character == '^':
                left_side.append(character)

        if character == '^' or len(right_side) > 0:
            if not character == '_':
                right_side.append(character)

    left_side_re = "".join(left_side[:-1])
    right_side_re = "".join(right_side[1:])

    ll = len(left_side_re)
    lr = len(right_side_re)

    forward = left_side_re + right_side_re
    reverse = reverse_complement(forward)

    forward = forward.replace('N', '[ACGT]')
    reverse = reverse.replace('N', '[ACGT]')

    return forward, ll, reverse, lr


def split_at_ligation_junction(sequence, pattern):
    patterns = []
    if isinstance(pattern, string_types):
        patterns.append(ligation_site_pattern(pattern))
    elif isinstance(pattern, tuple) or isinstance(pattern, list):
        if (isinstance(pattern, tuple) and len(pattern) == 4 and
                isinstance(pattern[1], int) and isinstance(pattern[3], int)):
            patterns.append(pattern)
        else:
            for p in pattern:
                if isinstance(p, tuple) and len(p) == 4:
                    patterns.append(p)
                else:
                    patterns.append(ligation_site_pattern(p))

    if isinstance(sequence, bytes):
        sequence = sequence.decode()

    hits = []
    for pattern in patterns:
        forward, lf, reverse, lr = pattern
        for m in re.finditer(forward, sequence, re.IGNORECASE):
            hits.append(m.start() + lf)

        if forward != reverse:
            for m in re.finditer(reverse, sequence, re.IGNORECASE):
                hits.append(m.start() + lr)

    sub_sequences = []
    previous_hit = 0
    for hit in sorted(hits):
        sub_sequences.append(sequence[previous_hit:hit])
        previous_hit = hit
    sub_sequences.append(sequence[previous_hit:])

    return sub_sequences


def add_dict(x, y):
    return {k: x.get(k, 0) + y.get(k, 0) for k in set(x) | set(y)}


class RareUpdateProgressBar(progressbar.ProgressBar):
    def __init__(self, min_value=0, max_value=None, widgets=None,
                 left_justify=True, initial_value=0, poll_interval=None,
                 percent_update_interval=1, silent=False, prefix=None,
                 **kwargs):
        self.manual_poll_interval = True
        if poll_interval is None:
            self.manual_poll_interval = False

        if silent:
            kwargs['fd'] = open(os.devnull, "w")

        if widgets is None:
            widgets = [progressbar.widgets.Percentage(), ' (', progressbar.widgets.SimpleProgress(), ') ',
                       progressbar.widgets.Bar(), ' ', progressbar.widgets.Timer(), ' ',
                       progressbar.widgets.AdaptiveETA()]

        if prefix is not None:
            widgets = [prefix + ' '] + widgets

        progressbar.ProgressBar.__init__(self, min_value=min_value, max_value=max_value, widgets=widgets,
                                         left_justify=left_justify, initial_value=initial_value,
                                         poll_interval=poll_interval, **kwargs)

        if self.max_value is not None:
            self.one_percent = self.max_value*(1.0*percent_update_interval)/100
        else:
            self.one_percent = 1

        self.silent = silent
        self.next_update = min_value

    def _needs_update(self):
        """
        Returns whether the ProgressBar should redraw the line.
        """
        if self.value > self.next_update or self.end_time:
            self.next_update += self.one_percent
            return True

        elif self.manual_poll_interval:
            delta = datetime.now() - self.last_update_time
            return delta > self.poll_interval


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


class WorkerMonitor(object):
    def __init__(self, value=0, manager=None):
        if manager is None:
            manager = mp.Manager()

        self.counter_lock = manager.Lock()
        self.worker_lock = manager.Lock()

        with self.counter_lock:
            self.val = value

        with self.worker_lock:
            self.worker_states = dict()

    def increment(self):
        with self.counter_lock:
            self.val += 1

    def value(self):
        with self.counter_lock:
            return self.val

    def set_worker_busy(self, worker_uuid):
        with self.worker_lock:
            self.worker_states[worker_uuid] = True

    def set_worker_idle(self, worker_uuid):
        with self.worker_lock:
            self.worker_states[worker_uuid] = False

    def get_worker_state(self, worker_uuid):
        with self.worker_lock:
            return self.worker_states[worker_uuid]

    def workers_idle(self):
        with self.worker_lock:
            for busy in self.worker_states.values():
                if busy:
                    return False
            return True


def get_sam_mapper(sam_file):
    try:
        if isinstance(sam_file, pysam.AlignmentFile):
            return sam_file.header['PG'][0]['ID']
        else:
            with pysam.AlignmentFile(sam_file) as sam:
                return sam.header['PG'][0]['ID']
    except Exception:
        return False



from ..pairs import ReadPairs
from .base_plotter import GenomeCoordFormatter
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import itertools


def summary_statistics_plot(stats, ax=None, **kwargs):
    """
    Barplot with filter statistics for FAN-C objects.

    :param stats: dictionary of the form {filter_name: int, ...}
    :param ax: (optional) axis in which to plot the bars.
               Will use :code:`plt.gca()` if not specified
    :param kwargs: Keyword options passed to Seaborn :code:`barplot`
    :return: Matplotlib axis
    """
    if ax is None:
        ax = plt.gca()

    palette = kwargs.pop('palette', 'colorblind')
    labels = []
    values = []

    if 'total' in stats:
        labels.append('total')
        values.append(stats['total'])

    if 'valid' in stats:
        labels.append('valid')
        values.append(stats['valid'])

    for key, value in sorted(stats.items()):
        if key in ('total', 'valid'):
            continue
        labels.append(key)
        values.append(value)

    sns.barplot(x=np.array(labels), y=np.array(values), ax=ax,
                palette=palette, **kwargs)
    sns.despine()

    ax.set_xticklabels([label.get_text()
                        for label in ax.get_xticklabels()],
                       rotation=90)

    ax.set_ylabel("Number of read pairs")
    ax.ticklabel_format(axis='y', style='sci', scilimits=(-2, 2))
    ax.figure.tight_layout()

    return ax


def ligation_bias_plot(pairs, ax=None, log=False, **kwargs):
    """
    Plot the ligation error structure of a dataset.

    :param pairs: Read pairs mapped to genomic regions (:class:`~fanc.pairs.ReadPairs`)
    :param ax: (optional) axis in which to plot the bars.
               Will use :code:`plt.gca()` if not specified
    :param log: log2-transform ratios if True
    :param kwargs: Additional arguments to pass
                   to :code:`plot`
    """
    if ax is None:
        ax = plt.gca()

    if isinstance(pairs, ReadPairs):
        x, inward_ratios, outward_ratios, bins_sizes = pairs.get_ligation_structure_biases()
    else:
        x, inward_ratios, outward_ratios, bins_sizes = pairs

    if log:
        inward_ratios = np.log2(inward_ratios) + 1
        outward_ratios = np.log2(outward_ratios) + 1

    with sns.axes_style("white", {
            "legend.frameon": True,
            "xtick.major.size": 4,
            "xtick.minor.size": 2,
            "ytick.major.size": 4,
            "ytick.minor.size": 2,
            "axes.linewidth": 0.5
    }):
        ax.set_title("Error structure by distance")
        ax.plot(x, inward_ratios, 'b', label="inward/same strand", **kwargs)
        ax.plot(x, outward_ratios, 'r', label="outward/same strand", **kwargs)
        ax.set_xscale('log')
        if log:
            ax.axhline(y=0, color='black', ls='dashed', lw=0.8)
            ax.set_ylim((-3, 3))
        else:
            ax.axhline(y=0.5, color='black', ls='dashed', lw=0.8)
            ax.set_ylim((0, 3))
        ax.set_xlabel('Gap size between fragments')
        ax.set_ylabel('Read count ratio')
        ax.legend(loc='upper right')
        sns.despine()

    ax.figure.tight_layout()
    return ax


def restriction_site_distance_plot(pairs, ax=None, max_percentile=95,
                                   sample=100000, max_distance=None,
                                   **kwargs):
    """
    Plot the distribution of read pair restriction site distances.

    The sum of distances to the nearest restriction site of each mate
    is equivalent to the insert size of the sequenced DNA molecule. As
    such, this plot can serve as a quality control for DNA fragmentation
    prior to sequencing. It can also be used to derive a cutoff for the
    :class:`~fanc.pairs.ReDistanceFilter`, which excludes mate pairs with
    very large insert sizes.

    :param pairs: :class:`~fanc.pairs.ReadPairs`
    :param ax: (optional) matplotlib axis
    :param max_percentile: Percentile of values up to which the distribution
                           is plotted. If this is set to 100, the distribution
                           will be squeezed in a small section on
                           the x axis if there are extremely large values
                           in the distribution. The default (95) makes sure
                           the distribution is properly visible.

    :param sample: If this is an integer, only <sample> random number of mate
                   pairs are plotted in the distribution to save time. If this
                   is set to None, all mate pairs are plotted. By default, a
                   sample of 100000 is shown.
    :param max_distance: If this is different from None, distances larger
                         than <max_distance> are ignored.
    :param kwargs: Keyword arguments passed to :code:`seaborn.distplot`
    :return: ax
    """
    if ax is None:
        ax = plt.gca()

    color = kwargs.pop('color', '#FBAFE4')

    distances = []
    for i, pair in enumerate(pairs.pairs(lazy=True)):
        d = pair.left.re_distance() + pair.right.re_distance()

        if max_distance is None or d <= max_distance:
            distances.append(d)
        if sample is not None and i >= sample:
            break

    distances = np.array(distances)
    if max_percentile is not None:
        median_insert, high = np.nanpercentile(distances, [50, max_percentile])
        distances = distances[distances < high]
        ax.set_xlim((0, high))
    else:
        median_insert = np.nanmedian(distances)
        ax.set_xlim(left=0)

    sns.distplot(distances, ax=ax, color=color, **kwargs)
    ax.axvline(x=median_insert, color='grey', linestyle='--')
    ax.set_xlabel("Sum of restriction site distances\n(insert size)")
    ax.set_ylabel("Density")
    sns.despine(ax=ax)
    ax.figure.tight_layout()

    return ax


def marginals_plot(matrix, chromosome, ax=None, lower=None, rel_cutoff=0.1, color='#3E896D',
                   **kwargs):
    """
    Plot Hi-C marginals vector.

    Marginals are the sum of values in each column of the matrix.
    This plot can be used to determine sensible low coverage thresholds
    for the :class:`~fanc.hic.LowCoverageFilter`.

    :param matrix: :class:`~fanc.matrix.RegionMatrixContainer`
    :param chromosome: Name of a chromosome to plot marginals for.
    :param ax: Matplotlib axis
    :param lower: Absolute lower cutoff for drawing threshold line
    :param rel_cutoff: Relative lower cutoff for drawing thresold line
    :param color: Color of line in plot
    :param kwargs: Keyword arguments passed to
                   :func:`~fanc.matrix.RegionMatrixContainer.matrix`
    :return: ax
    """
    if ax is None:
        ax = plt.gca()

    bin_size = matrix.bin_size

    kwargs['key'] = (chromosome, chromosome)
    marginals = matrix.marginals(**kwargs)

    if lower is None and rel_cutoff is not None:
        median = np.nanmedian(marginals[marginals > 0])
        lower = rel_cutoff * median

    distances = np.arange(0, bin_size * len(marginals), bin_size)
    ax.plot(distances, marginals, color=color)

    if lower is not None:
        ax.axhline(lower, color='grey', linestyle='--')

    ax.xaxis.set_major_formatter(GenomeCoordFormatter(chromosome,
                                                      minor_div=5,
                                                      display_chromosome=True,
                                                      display_scale=False))
    ax.set_ylabel("Sum of counts")
    ax.ticklabel_format(axis='y', style='sci', scilimits=(-2, 2))
    ax.set_ylim(bottom=0)
    sns.despine(ax=ax)
    ax.figure.tight_layout()

    return ax


def distance_decay_plot(*matrices, ax=None, chromosome=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    for matrix in matrices:
        ex, ex_chromosome, ex_inter = matrix.expected_values()

        if chromosome is not None:
            ex = ex_chromosome[chromosome]

        bin_size = matrix.bin_size
        distances = np.arange(0, bin_size * len(ex), bin_size)
        ax.plot(distances, ex, **kwargs)

    ax.set_xscale('log')
    ax.set_yscale('log')

    ax.set_ylabel('Expected value')
    ax.set_xlabel('Genomic separation')
    ax.xaxis.set_major_formatter(GenomeCoordFormatter(chromosome if chromosome is not None else "All",
                                                      minor_div=5,
                                                      display_chromosome=False,
                                                      display_scale=False))
    ax.figure.tight_layout()
    return ax


def pca_plot(pca_res, variance=None, eigenvectors=(0, 1),
             markers=None, colors=None, names=None, ax=None):
    if markers is None:
        markers = ('^', 'o', '*', 's', 'D', 'v', 'd', 'H', 'p', '>')
    if colors is None:
        colors = ('red', 'blue', 'green', 'purple', 'yellow', 'black',
                  'orange', 'pink', 'cyan', 'lawngreen')
    markers = itertools.cycle(markers)
    colors = itertools.cycle(colors)

    xlabel = 'PC1'
    if variance is not None:
        xlabel += ' (%d%%)' % int(variance[eigenvectors[0]]*100)

    ylabel = 'PC2'
    if variance is not None:
        ylabel += ' (%d%%)' % int(variance[eigenvectors[1]]*100)

    if ax is None:
        if names is not None:
            ax = plt.subplot(121)
        else:
            ax = plt.subplot(111)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    ax.set_title('PCA on %d samples' % pca_res.shape[0])

    for i in range(pca_res.shape[0]):
        name = names[i] if names is not None else None
        ax.plot(pca_res[i, eigenvectors[0]], pca_res[i, eigenvectors[1]],
                marker=next(markers), color=next(colors), label=name)

    if names is not None:
        ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    return ax.figure, ax




import warnings
import numpy as np
from fanc.tools.matrix import remove_sparse_rows, restore_sparse_rows
from fanc.legacy.data.genomic import Hic, AccessOptimisedHic, Edge
import logging
logger = logging.getLogger(__name__)


def correct(hic, whole_matrix=True, intra_chromosomal=True, inter_chromosomal=True,
            copy=False, file_name=None, optimise=True,
            restore_coverage=False):
    hic_new = None
    chromosome_starts = dict()
    last_chromosome = None
    if copy:
        if optimise:
            hic_new = AccessOptimisedHic(file_name=file_name, mode='w')
        else:
            hic_new = Hic(file_name=file_name, mode='w')

        for i, region in enumerate(hic.regions()):
            hic_new.add_region(region, flush=False)
            if region.chromosome != last_chromosome:
                chromosome_starts[region.chromosome] = i
            last_chromosome = region.chromosome
        hic_new.flush()
        hic_new.disable_indexes()

    if not whole_matrix:
        bias_vectors = []
        for chromosome in hic.chromosomes():
            m = hic[chromosome, chromosome]
            m_corrected, bias_vector_chromosome = correct_matrix(m, restore_coverage=restore_coverage)
            if hic_new is None:
                logger.debug("Replacing corrected edges in existing Hic object...")
                hic[chromosome, chromosome] = m_corrected
            else:
                logger.debug("Adding corrected edges in new Hic object ...")
                chromosome_offset = chromosome_starts[chromosome]
                for i in range(m_corrected.shape[0]):
                    i_region = i + chromosome_offset
                    nonzero_idx = np.nonzero(m_corrected[i])[0]
                    for j in nonzero_idx[nonzero_idx >= i]:
                        j_region = j + chromosome_offset
                        weight = m_corrected[i, j]
                        hic_new.add_edge(Edge(source=i_region, sink=j_region, weight=weight), flush=False)
            bias_vectors.append(bias_vector_chromosome)
        logger.debug("Done.")
        logger.debug("Adding bias vector...")
        if hic_new is None:
            hic.bias_vector(np.concatenate(bias_vectors))
        else:
            hic_new.bias_vector(np.concatenate(bias_vectors))
        logger.debug("Done.")
    else:
        logger.debug("Fetching whole genome matrix")
        m = hic[:, :]
        cb = hic.chromosome_bins
        if not intra_chromosomal:
            for chromosome, bins in cb.items():
                m[bins[0]:bins[1], bins[0]:bins[1]] = 0
        if not inter_chromosomal:
            for chromosome, bins in cb:
                m[0:bins[0], bins[0]:bins[1]] = 0
                m[bins[1]:, bins[0]:bins[1]] = 0

        m_corrected, bias_vector = correct_matrix(m, restore_coverage=restore_coverage)
        if hic_new is None:
            logger.debug("Replacing corrected edges in existing Hic object...")
            hic[:, :] = m_corrected
        else:
            logger.debug("Adding corrected edges in new Hic object ...")
            for i in range(m_corrected.shape[0]):
                nonzero_idx = np.nonzero(m_corrected[i])[0]
                for j in nonzero_idx[nonzero_idx >= i]:
                    weight = m_corrected[i, j]
                    hic_new.add_edge([i, j, weight], flush=False)
        logger.debug("Done.")
        logger.debug("Adding bias vector...")
        if hic_new is None:
            hic.bias_vector(bias_vector)
        else:
            hic_new.bias_vector(bias_vector)
        logger.debug("Done.")

    if hic_new is None:
        return hic
    hic_new.flush()
    hic_new.enable_indexes()
    return hic_new


def correct_matrix(m, max_attempts=50, restore_coverage=False):
    # remove zero-sum rows
    removed_rows = []
    m_nonzero, ixs = remove_sparse_rows(m, cutoff=0)
    removed_rows.append(ixs)

    has_errors = True
    iterations = 0
    x = None
    while has_errors:
        has_errors = False

        try:
            x = get_bias_vector(m_nonzero)
        except ValueError as e:
            logger.debug("Matrix balancing failed (this can happen!), \
                          removing sparsest rows to try again. Error: \
                          %s" % str(e))
            m_nonzero, ixs = remove_sparse_rows(m_nonzero)
            removed_rows.append(ixs)
            has_errors = True

        iterations += 1
        if iterations > max_attempts:
            raise RuntimeError("Exceeded maximum attempts (%d)" % max_attempts)

    if restore_coverage:
        x = x*np.sqrt(np.sum(m_nonzero)/m_nonzero.shape[0])

    logger.debug("Applying bias vector")
    m_nonzero = x*m_nonzero*x[:, np.newaxis]

    logger.debug(removed_rows)
    logger.debug("Restoring {} sets ({} total) sparse rows.".format(
        len(removed_rows), sum(len(x) for x in removed_rows)))
    # restore zero rows
    m_nonzero = restore_sparse_rows(m_nonzero, removed_rows)
    x = restore_sparse_rows(x, removed_rows)

    return m_nonzero, x


def get_bias_vector(A, x0=None, tol=1e-06, delta=0.1, Delta=3, fl=0, high_precision=False, outer_limit=300):
    logger.debug("Starting matrix balancing")

    with warnings.catch_warnings():
        warnings.filterwarnings('error')

        try:
            # basic variables
            # n=size_(A,1)
            if not isinstance(A, np.ndarray):
                try:
                    if high_precision:
                        A = np.array(A, dtype=np.float128)
                    else:
                        A = np.array(A)
                except AttributeError:
                    A = np.array(A)
            n = A.shape[0]
            # e=ones_(n,1)
            try:
                if high_precision:
                    e = np.ones(n, dtype=np.float128)
                else:
                    e = np.ones(n)
            except AttributeError:
                e = np.ones(n)

            if not x0:
                try:
                    if high_precision:
                        x0 = np.ones(n, np.float128)
                    else:
                        x0 = np.ones(n)
                except AttributeError:
                    x0 = np.ones(n)
            else:
                try:
                    if high_precision:
                        x0 = np.array(x0, np.float128)
                    else:
                        x0 = np.array(x0)
                except AttributeError:
                    x0 = np.array(x0)
            res = np.array([])
            g = 0.9
            etamax = 0.1
            eta = etamax
            stop_tol = tol * 0.5
            x = x0.copy()
            rt = tol ** 2
            # v=x.dot((A * x))
            v = x*A.dot(x)
            rk = 1 - v
            # rho_km1=rk.T * rk
            rho_km1 = rk.T.dot(rk)
            rout = rho_km1
            rho_km2 = rho_km1
            rold = rout
            MVP = 0
            i = 0

            n_iterations_outer = 0
            while rout > rt:
                n_iterations_outer += 1

                if n_iterations_outer > outer_limit:
                    raise ValueError("Number of iterations has exceeded the limit (%d)." % outer_limit)

                i += 1
                k = 0
                y = e.copy()
                innertol = max(eta ** 2 * rout, rt)
                n_iterations_inner = 0
                while rho_km1 > innertol:
                    n_iterations_inner += 1

                    k += 1
                    if k == 1:
                        try:
                            Z = rk / v
                        except Warning:
                            raise ValueError("v=0; Remove zero or sparse rows")
                        p = Z.copy()
                        rho_km1 = rk.T.dot(Z)
                    else:
                        beta = rho_km1 / rho_km2
                        p = Z + beta * p
                    # w = x.*(A*(x.*p)) + v.*p;
                    w = x*A.dot(x*p) + v*p
                    alpha = rho_km1 / p.T.dot(w)
                    ap = alpha * p
                    ynew = y + ap
                    if min(ynew) <= delta:
                        if delta == 0:
                            break
                        ind = np.where(ap < 0)[0]
                        # gamma = min((delta  - y(ind))./ap(ind));
                        gamma = min((delta-y[ind])/ap[ind])
                        y = y + gamma * ap
                        break
                    if max(ynew) >= Delta:
                        ind = np.where(ynew > Delta)[0]
                        gamma = min((Delta-y[ind])/ap[ind])
                        y = y + gamma * ap
                        break
                    y = ynew.copy()
                    rk = rk - alpha * w
                    rho_km2 = rho_km1.copy()
                    Z = rk / v
                    rho_km1 = rk.T.dot(Z)

                try:
                    x = x*y
                except Warning:
                    raise ValueError("Value in x or y too small to represent numerically. Try removing sparse rows")

                v = x*A.dot(x)
                rk = 1 - v
                rho_km1 = rk.T.dot(rk)
                rout = rho_km1.copy()
                MVP = MVP + k + 1
                rat = rout / rold
                rold = rout.copy()
                res_norm = np.sqrt(rout)
                eta_o = eta
                eta = g * rat
                if g * eta_o ** 2 > 0.1:
                    eta = max(eta, g * eta_o ** 2)
                eta = max(min(eta, etamax), stop_tol / res_norm)
                if fl == 1:
                    res = np.array([[res], [res_norm]])

            logger.debug("Matrix-vector products = %d\n" % MVP)
            logger.debug("Outer iterations: %d" % n_iterations_outer)
        except Warning as e:
            logger.error(str(e))
            raise ValueError("Generic catch all warnings")

    return x

